"""DWD rules."""

import dataclasses
import datetime
import logging
import re
import typing

import HABApp

import habapp_rules.actors.state_observer
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.dwd

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class DwdItems:
    """Dataclass for DWD items needed by DwdWindAlarm."""

    description: HABApp.openhab.items.StringItem
    warn_type: HABApp.openhab.items.StringItem
    severity: HABApp.openhab.items.StringItem
    start_time: HABApp.openhab.items.DatetimeItem
    end_time: HABApp.openhab.items.DatetimeItem

    SEVERITY_MAPPING: typing.ClassVar = {"NULL": 0, "Minor": 1, "Moderate": 2, "Severe": 3, "Extreme": 4}

    @classmethod
    def from_prefix(cls, prefix: str) -> "DwdItems":
        """Init DwdItems from prefix.

        Args:
            prefix: common prefix of all DWD items

        Returns:
            Dataclass with all needed DWD items
        """
        description = HABApp.openhab.items.StringItem.get_item(f"{prefix}_description")
        warn_type = HABApp.openhab.items.StringItem.get_item(f"{prefix}_type")
        severity = HABApp.openhab.items.StringItem.get_item(f"{prefix}_severity")
        start_time = HABApp.openhab.items.DatetimeItem.get_item(f"{prefix}_start_time")
        end_time = HABApp.openhab.items.DatetimeItem.get_item(f"{prefix}_end_time")

        return cls(description, warn_type, severity, start_time, end_time)

    @property
    def severity_as_int(self) -> int:
        """Get severity as integer.

        Returns:
            severity as integer value
        """
        return self.SEVERITY_MAPPING.get(self.severity.value, 0)


class DwdWindAlarm(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Rule for setting wind alarm by DWD warnings.

    # Items:
    Switch      I26_99_wind_alarm               "Wind alarm"
    Switch      I26_99_wind_alarm_manual        "Wind alarm manual"
    String      I26_99_wind_alarm_state         "Wind alarm state"

    String      I26_99_warning_1_severity       "Severity"              {channel="dwdunwetter:dwdwarnings:ingolstadt:severity1"}
    String      I26_99_warning_1_description    "Description"           {channel="dwdunwetter:dwdwarnings:ingolstadt:description1"}
    DateTime    I26_99_warning_1_start_time     "valid from"            {channel="dwdunwetter:dwdwarnings:ingolstadt:onset1"}
    DateTime    I26_99_warning_1_end_time       "valid till"            {channel="dwdunwetter:dwdwarnings:ingolstadt:expires1"}
    String      I26_99_warning_1_type           "Type"                  {channel="dwdunwetter:dwdwarnings:ingolstadt:event1"}

    String      I26_99_warning_2_severity       "Severity"              {channel="dwdunwetter:dwdwarnings:ingolstadt:severity2"}
    String      I26_99_warning_2_description    "Description"           {channel="dwdunwetter:dwdwarnings:ingolstadt:description2"}
    DateTime    I26_99_warning_2_start_time     "valid from"            {channel="dwdunwetter:dwdwarnings:ingolstadt:onset2"}
    DateTime    I26_99_warning_2_end_time       "valid till"            {channel="dwdunwetter:dwdwarnings:ingolstadt:expires2"}
    String      I26_99_warning_2_type           "Type"                  {channel="dwdunwetter:dwdwarnings:ingolstadt:event2"}

    # Config
    config = habapp_rules.sensors.config.dwd.WindAlarmConfig(
            items=habapp_rules.sensors.config.dwd.WindAlarmItems(
                    wind_alarm="I26_99_wind_alarm",
                    manual="I26_99_wind_alarm_manual",
                    state="I26_99_wind_alarm_state"
            ),
            parameter=habapp_rules.sensors.config.dwd.WindAlarmParameter(

            )
    )

    # Rule init:
    habapp_rules.sensors.dwd.DwdWindAlarm(config)
    """

    states: typing.ClassVar = [{"name": "Manual"}, {"name": "Hand", "timeout": 20 * 3600, "on_timeout": "_auto_hand_timeout"}, {"name": "Auto", "initial": "Init", "children": [{"name": "Init"}, {"name": "On"}, {"name": "Off"}]}]

    trans: typing.ClassVar = [
        {"trigger": "manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
        {"trigger": "manual_off", "source": "Manual", "dest": "Auto"},
        {"trigger": "hand", "source": "Auto", "dest": "Hand"},
        {"trigger": "wind_alarm_start", "source": "Auto_Off", "dest": "Auto_On"},
        {"trigger": "wind_alarm_end", "source": "Auto_On", "dest": "Auto_Off"},
    ]

    def __init__(self, config: habapp_rules.sensors.config.dwd.WindAlarmConfig) -> None:
        """Init of DWD wind alarm object.

        Args:
            config: config for DWD wind alarm rule

        Raises:
            TypeError: if type of hand_timeout is not supported
        """
        self._config = config
        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.wind_alarm.name)

        self._items_dwd = [DwdItems.from_prefix(f"{self._config.parameter.dwd_item_prefix}{idx + 1}") for idx in range(self._config.parameter.number_dwd_objects)]

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
            model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state"
        )  # this function is missing!

        self._wind_alarm_observer = habapp_rules.actors.state_observer.StateObserverSwitch(self._config.items.wind_alarm.name, self._cb_hand, self._cb_hand)

        self._set_timeouts()
        self._set_initial_state()

        # callbacks
        self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.hand_timeout is not None:
            self._config.items.hand_timeout.listen_event(self._cb_hand_timeout, HABApp.openhab.events.ItemStateChangedEventFilter())

        self.run.at(self.run.trigger.interval(None, 300), self._cb_cyclic_check)

    def _get_initial_state(self, default_value: str = "") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._config.items.manual.is_on():
            return "Manual"
        if self._wind_alarm_active():
            return "Auto_On"
        return "Auto_Off"

    def _get_hand_timeout(self) -> int:
        """Get value of hand timeout.

        Returns:
            hand timeout in seconds (0 is no timeout)
        """
        if self._config.items.hand_timeout is not None:
            if (item_value := self._config.items.hand_timeout.value) is None:
                self._instance_logger.warning("The value of the hand timeout item is None. Will use 24 hours as default!")
                return 24 * 3600
            return item_value
        return self._config.parameter.hand_timeout

    def _set_timeouts(self) -> None:
        """Set timeouts."""
        self.state_machine.get_state("Hand").timeout = self._get_hand_timeout()

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This should method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            if self.state == "Auto_On" and self._wind_alarm_observer.value is not True:
                self._wind_alarm_observer.send_command("ON")
            elif self.state == "Auto_Off" and self._wind_alarm_observer.value is not False:
                self._wind_alarm_observer.send_command("OFF")

            self._previous_state = self.state

    def on_enter_Auto_Init(self) -> None:  # noqa: N802
        """Is called on entering of init state."""
        self._set_initial_state()

    def _cb_hand(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:  # noqa: ARG002
        """Callback, which is triggered by the state observer if a manual change was detected.

        Args:
            event: original trigger event
        """
        self.hand()

    def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the manual switch has a state change event.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.manual_on()
        else:
            self.manual_off()

    def _cb_hand_timeout(self, _: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the timeout item changed."""
        self._set_timeouts()

    def _wind_alarm_active(self) -> bool:
        """Check if wind alarm is active.

        Returns:
            True if active, False if not
        """
        for dwd_items in self._items_dwd:
            if dwd_items.warn_type.value in {"BÖEN", "WIND", "STURM", "GEWITTER"}:
                speed_values = [int(value) for value in re.findall(r"\b(\d+)\s*km/h\b", dwd_items.description.value)]
                if not speed_values:
                    continue

                if max(speed_values) >= self._config.parameter.threshold_wind_speed and dwd_items.severity_as_int >= self._config.parameter.threshold_severity and dwd_items.start_time.value < datetime.datetime.now() < dwd_items.end_time.value:
                    return True
        return False

    def _cb_cyclic_check(self) -> None:
        """Callback to check if wind alarm is active. This should be called cyclic every few minutes."""
        if self.state not in {"Auto_On", "Auto_Off"}:
            return

        if self._wind_alarm_active() and self.state != "Auto_On":
            self.wind_alarm_start()

        elif not self._wind_alarm_active() and self.state != "Auto_Off":
            self.wind_alarm_end()



"""Rules for notification."""

import HABApp
from multi_notifier.connectors.connector_telegram import Telegram

import habapp_rules.system.config.notification


class SendStateChanged(HABApp.Rule):
    """Rule class to send a telegram if the state of an item changes."""

    def __init__(self, config: habapp_rules.system.config.notification.NotificationConfig) -> None:
        """Init the rule object.

        Args:
            config: config for notification rule
        """
        self._config = config
        HABApp.Rule.__init__(self)

        self._config.items.target_item.listen_event(self._send_state_change, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _send_state_change(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is called if the state of the item changed.

        Args:
            event: event which triggered the callback
        """
        msg = f"{event.name} changed from {event.old_value} to {event.value}"

        if isinstance(self._config.parameter.notify_connector, Telegram):
            self._config.parameter.notify_connector.send_message(self._config.parameter.recipients, msg)
        else:
            self._config.parameter.notify_connector.send_message(self._config.parameter.recipients, msg, subject=f"{event.name} changed")



"""Test pydantic base models."""

import unittest.mock

import HABApp
import pydantic

import habapp_rules.core.exceptions
import habapp_rules.core.pydantic_base
import tests.helper.oh_item
import tests.helper.test_case_base


class ItemsForTesting(habapp_rules.core.pydantic_base.ItemBase):
    """Items for testing."""

    switch: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item for testing")
    switch_create: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item for testing", json_schema_extra={"create_if_not_exists": True})
    dimmer_list: list[HABApp.openhab.items.DimmerItem] = pydantic.Field(..., description="list of dimmer items for testing")
    optional_contact: HABApp.openhab.items.ContactItem | None = pydantic.Field(None, description="optional contact item for testing")
    not_supported: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="not supported item for testing")


class ItemsListCreateException(habapp_rules.core.pydantic_base.ItemBase):
    """Model with list object where create_if_not_exists is set."""

    some_items: list[HABApp.openhab.items.SwitchItem | HABApp.openhab.items.DimmerItem] = pydantic.Field(..., description="list of items for testing", json_schema_extra={"create_if_not_exists": True})


class WrongTypeException(habapp_rules.core.pydantic_base.ItemBase):
    """Model with wrong type."""

    item: str = pydantic.Field(..., description="wrong type for testing")


class MultipleTypeForCreateException(habapp_rules.core.pydantic_base.ItemBase):
    """Model with multiple types where create_if_not_exists is set."""

    item: HABApp.openhab.items.SwitchItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="list of items for testing", json_schema_extra={"create_if_not_exists": True})


class TestItemBase(tests.helper.test_case_base.TestCaseBase):
    """Test ItemBase."""

    def test_check_all_fields_oh_items_exceptions(self) -> None:
        """Test all exceptions of check_all_fields_oh_items."""
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            ItemsListCreateException(some_items=["Name1", "Name2"])

        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            WrongTypeException(item="Name1")

        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            MultipleTypeForCreateException(item="Name1")

    def test_convert_to_oh_item(self) -> None:
        """Test convert_to_oh_item."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_2", None)

        dimmer = HABApp.openhab.items.DimmerItem.get_item("Unittest_Dimmer_2")

        with (
            unittest.mock.patch("habapp_rules.core.helper.create_additional_item", return_value=HABApp.openhab.items.SwitchItem("Unittest_Switch_Created", "")) as create_item_mock,
            self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError),
        ):
            ItemsForTesting(
                switch="Unittest_Switch",  # normal case
                switch_create="Unittest_Switch_Created",  # item which will be created
                dimmer_list=["Unittest_Dimmer_1", dimmer],  # mixed list of strings and HABApp.openhab.items.DimmerItem
                optional_contact=None,  # test if None is OK
                not_supported=5,  # this causes an exception
            )

        create_item_mock.assert_called_once_with("Unittest_Switch_Created", "Switch")



"""Test energy save switch rules."""

import collections
import pathlib
import sys
import unittest.mock

import HABApp.rule.rule

import habapp_rules.actors.config.energy_save_switch
import habapp_rules.actors.config.shading
import habapp_rules.actors.energy_save_switch
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer
from habapp_rules.system import PresenceState, SleepState


class TestEnergySaveSwitch(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing energy save switch."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Min_Switch")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Min_State")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Max_Switch")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Max_State")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Max_Manual")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_External_Request")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Current_Switch")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Current_State")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Current_Manual")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Current")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state")

        self._config_min = habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchConfig(items=habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchItems(switch="Unittest_Min_Switch", state="Unittest_Min_State"))

        self._config_max_without_current = habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchConfig(
            items=habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchItems(
                switch="Unittest_Max_Switch", state="Unittest_Max_State", manual="Unittest_Max_Manual", external_request="Unittest_External_Request", presence_state="Unittest_Presence_state", sleeping_state="Unittest_Sleep_state"
            ),
            parameter=habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchParameter(max_on_time=3600, hand_timeout=1800),
        )

        self._config_current = habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchConfig(
            items=habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchItems(
                switch="Unittest_Current_Switch", state="Unittest_Current_State", manual="Unittest_Current_Manual", current="Unittest_Current", presence_state="Unittest_Presence_state", sleeping_state="Unittest_Sleep_state"
            ),
            parameter=habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchParameter(current_threshold=0.1, extended_wait_for_current_time=142),
        )

        self._rule_min = habapp_rules.actors.energy_save_switch.EnergySaveSwitch(self._config_min)
        self._rule_max_without_current = habapp_rules.actors.energy_save_switch.EnergySaveSwitch(self._config_max_without_current)
        self._rule_with_current = habapp_rules.actors.energy_save_switch.EnergySaveSwitch(self._config_current)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "EnergySaveSwitch"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self._rule_min.states, transitions=self._rule_min.trans, initial=self._rule_min.state, show_conditions=False)

        graph.get_graph().draw(picture_dir / "EnergySaveSwitch.png", format="png", prog="dot")

        for state_name in [state for state in self._get_state_names(self._rule_min.states) if "init" not in state.lower()]:
            graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self._rule_min.states, transitions=self._rule_min.trans, initial=state_name, show_conditions=True)
            graph.get_graph(force_new=True, show_roi=True).draw(picture_dir / f"EnergySaveSwitch_{state_name}.png", format="png", prog="dot")

    def test_set_timeout(self) -> None:
        """Test set timeout."""
        self.assertEqual(self._rule_min.state_machine.states["Hand"].timeout, 0)
        self.assertEqual(self._rule_max_without_current.state_machine.states["Hand"].timeout, 1800)
        self.assertEqual(self._rule_with_current.state_machine.states["Hand"].timeout, 0)

        self.assertEqual(self._rule_min.state_machine.states["Auto"].states["WaitCurrentExtended"].timeout, 60)
        self.assertEqual(self._rule_max_without_current.state_machine.states["Auto"].states["WaitCurrentExtended"].timeout, 60)
        self.assertEqual(self._rule_with_current.state_machine.states["Auto"].states["WaitCurrentExtended"].timeout, 142)

    def test_get_initial_state(self) -> None:
        """Test get initial state."""
        TestCase = collections.namedtuple("TestCase", "current_above_threshold, manual, on_conditions_met, expected_state")

        test_cases = [
            # current below threshold
            TestCase(current_above_threshold=False, manual=None, on_conditions_met=False, expected_state="Auto_Off"),
            TestCase(current_above_threshold=False, manual=None, on_conditions_met=True, expected_state="Auto_On"),
            TestCase(current_above_threshold=False, manual=False, on_conditions_met=False, expected_state="Auto_Off"),
            TestCase(current_above_threshold=False, manual=False, on_conditions_met=True, expected_state="Auto_On"),
            TestCase(current_above_threshold=False, manual=True, on_conditions_met=False, expected_state="Manual"),
            TestCase(current_above_threshold=False, manual=True, on_conditions_met=True, expected_state="Manual"),
            # current above threshold
            TestCase(current_above_threshold=True, manual=None, on_conditions_met=False, expected_state="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, manual=None, on_conditions_met=True, expected_state="Auto_On"),
            TestCase(current_above_threshold=True, manual=False, on_conditions_met=False, expected_state="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, manual=False, on_conditions_met=True, expected_state="Auto_On"),
            TestCase(current_above_threshold=True, manual=True, on_conditions_met=False, expected_state="Manual"),
            TestCase(current_above_threshold=True, manual=True, on_conditions_met=True, expected_state="Manual"),
        ]

        with unittest.mock.patch.object(self._rule_max_without_current, "_get_on_off_conditions_met") as on_conditions_mock, unittest.mock.patch.object(self._rule_max_without_current, "_current_above_threshold") as current_above_threshold_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    on_conditions_mock.return_value = test_case.on_conditions_met
                    current_above_threshold_mock.return_value = test_case.current_above_threshold

                    if test_case.manual is None:
                        self._rule_max_without_current._config.items.manual = None
                    else:
                        self._rule_max_without_current._config.items.manual = unittest.mock.MagicMock()
                        self._rule_max_without_current._config.items.manual.is_on.return_value = test_case.manual

                    self.assertEqual(test_case.expected_state, self._rule_max_without_current._get_initial_state())

    def test_current_above_threshold(self) -> None:
        """Test current above threshold."""
        TestCase = collections.namedtuple("TestCase", "current, threshold, expected_result")

        test_cases = [
            TestCase(current=None, threshold=0.1, expected_result=False),
            TestCase(current=None, threshold=0.1, expected_result=False),
            TestCase(current=None, threshold=0.1, expected_result=False),
            TestCase(current=0.0, threshold=0.1, expected_result=False),
            TestCase(current=0.1, threshold=0.1, expected_result=False),
            TestCase(current=0.2, threshold=0.1, expected_result=True),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                if test_case.current is None:
                    self._rule_with_current._config.items.current = None
                else:
                    self._rule_with_current._config.items.current = unittest.mock.MagicMock(value=test_case.current)

                self._rule_with_current._config.parameter.current_threshold = test_case.threshold
                self.assertEqual(test_case.expected_result, self._rule_with_current._current_above_threshold())

    def test_auto_off_transitions(self) -> None:
        """Test auto off transitions."""
        TestCase = collections.namedtuple("TestCase", "external_req, sleeping_state, presence_state, expected_state")
        test_cases = [
            TestCase(external_req="OFF", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.ABSENCE, expected_state="Auto_Off"),
            TestCase(external_req="OFF", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.PRESENCE, expected_state="Auto_Off"),
            TestCase(external_req="OFF", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.ABSENCE, expected_state="Auto_Off"),
            TestCase(external_req="OFF", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.PRESENCE, expected_state="Auto_On"),
            TestCase(external_req="ON", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.ABSENCE, expected_state="Auto_On"),
            TestCase(external_req="ON", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.PRESENCE, expected_state="Auto_On"),
            TestCase(external_req="ON", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.ABSENCE, expected_state="Auto_On"),
            TestCase(external_req="ON", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.PRESENCE, expected_state="Auto_On"),
        ]

        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Min_State", "Auto_Off")

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.item_state_change_event("Unittest_External_Request", test_case.external_req)
                tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", test_case.sleeping_state.value)
                tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", test_case.presence_state.value)

                tests.helper.oh_item.assert_value("Unittest_Max_State", test_case.expected_state)
                tests.helper.oh_item.assert_value("Unittest_Max_Switch", "ON" if test_case.expected_state == "Auto_On" else "OFF")

                tests.helper.oh_item.assert_value("Unittest_Min_State", "Auto_Off")

    def test_auto_on_transitions(self) -> None:
        """Test auto on transitions."""
        # max on time
        self._rule_min.to_Auto_On()
        self._rule_max_without_current.to_Auto_On()
        self._rule_with_current.to_Auto_On()

        self.assertIsNone(self._rule_min._max_on_countdown)
        self.assertIsNotNone(self._rule_max_without_current._max_on_countdown)
        self.assertIsNone(self._rule_with_current._max_on_countdown)

        self._rule_min._cb_max_on_countdown()
        self._rule_max_without_current._cb_max_on_countdown()
        self._rule_with_current._cb_max_on_countdown()

        tests.helper.oh_item.assert_value("Unittest_Min_State", "Auto_On")
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_On")

        # off conditions met
        TestCase = collections.namedtuple("TestCase", "current_above_threshold, external_req, sleeping_state, presence_state, expected_state_max, expected_state_current")
        test_cases = [
            TestCase(current_above_threshold=False, external_req="OFF", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_Off", expected_state_current="Auto_Off"),
            TestCase(current_above_threshold=False, external_req="OFF", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_Off", expected_state_current="Auto_Off"),
            TestCase(current_above_threshold=False, external_req="OFF", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_Off", expected_state_current="Auto_Off"),
            TestCase(current_above_threshold=False, external_req="OFF", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_On", expected_state_current="Auto_On"),
            TestCase(current_above_threshold=False, external_req="ON", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_On", expected_state_current="Auto_Off"),
            TestCase(current_above_threshold=False, external_req="ON", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_On", expected_state_current="Auto_Off"),
            TestCase(current_above_threshold=False, external_req="ON", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_On", expected_state_current="Auto_Off"),
            TestCase(current_above_threshold=False, external_req="ON", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_On", expected_state_current="Auto_On"),
            TestCase(current_above_threshold=True, external_req="OFF", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_Off", expected_state_current="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, external_req="OFF", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_Off", expected_state_current="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, external_req="OFF", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_Off", expected_state_current="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, external_req="OFF", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_On", expected_state_current="Auto_On"),
            TestCase(current_above_threshold=True, external_req="ON", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_On", expected_state_current="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, external_req="ON", sleeping_state=SleepState.SLEEPING, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_On", expected_state_current="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, external_req="ON", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.ABSENCE, expected_state_max="Auto_On", expected_state_current="Auto_WaitCurrent"),
            TestCase(current_above_threshold=True, external_req="ON", sleeping_state=SleepState.AWAKE, presence_state=PresenceState.PRESENCE, expected_state_max="Auto_On", expected_state_current="Auto_On"),
        ]

        with unittest.mock.patch.object(self._rule_with_current, "_current_above_threshold", return_value=None) as mock_current_above_threshold:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    mock_current_above_threshold.return_value = test_case.current_above_threshold
                    self._rule_min.to_Auto_On()
                    self._rule_max_without_current.to_Auto_On()
                    self._rule_with_current.to_Auto_On()

                    tests.helper.oh_item.item_state_change_event("Unittest_External_Request", test_case.external_req)
                    tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", test_case.sleeping_state.value)
                    tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", test_case.presence_state.value)

                    tests.helper.oh_item.assert_value("Unittest_Max_State", test_case.expected_state_max)
                    tests.helper.oh_item.assert_value("Unittest_Max_Switch", "ON" if test_case.expected_state_max == "Auto_On" else "OFF")

                    tests.helper.oh_item.assert_value("Unittest_Current_State", test_case.expected_state_current)
                    tests.helper.oh_item.assert_value("Unittest_Current_Switch", "ON" if test_case.expected_state_current in {"Auto_On", "Auto_WaitCurrent"} else "OFF")

                    tests.helper.oh_item.assert_value("Unittest_Min_State", "Auto_On")

    def test_auto_on_transitions_timeout(self) -> None:
        """Test auto on transitions with max_on_countdown."""
        # external request OFF
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "OFF")
        self._rule_max_without_current.to_Auto_On()
        self._rule_max_without_current._cb_max_on_countdown()
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "OFF")

        # external request ON
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "ON")
        self._rule_max_without_current.to_Auto_On()
        self._rule_max_without_current._cb_max_on_countdown()
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_On")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "ON")
        # external to off
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "OFF")

    def test_auto_wait_current_transitions(self) -> None:
        """Test Auto_WaitCurrent transitions."""
        # on conditions met
        self._rule_with_current.to_Auto_WaitCurrent()
        self._rule_with_current.on_conditions_met()
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_On")
        tests.helper.oh_item.assert_value("Unittest_Current_Switch", "ON")

        # current below threshold
        self._rule_with_current.to_Auto_WaitCurrent()
        self._rule_with_current.current_below_threshold()
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_WaitCurrentExtended")
        tests.helper.oh_item.assert_value("Unittest_Current_Switch", "ON")

        # max_on_countdown | external request off
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "OFF")
        self._rule_max_without_current.to_Auto_WaitCurrent()
        self._rule_max_without_current._cb_max_on_countdown()
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "OFF")

        # max_on_countdown | external request on
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "ON")
        self._rule_max_without_current.to_Auto_WaitCurrent()
        self._rule_max_without_current._cb_max_on_countdown()
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_WaitCurrent")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "ON")
        # external to off
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "OFF")

    def test_hand_transitions(self) -> None:
        """Test Hand transitions."""
        # max_on_countdown | external request off
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "OFF")
        self._rule_max_without_current.to_Hand()
        self._rule_max_without_current._cb_max_on_countdown()
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "OFF")

        # max_on_countdown | external request on
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "ON")
        self._rule_max_without_current.to_Hand()
        self._rule_max_without_current._cb_max_on_countdown()
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Hand")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "ON")
        # external to off
        tests.helper.oh_item.item_state_change_event("Unittest_External_Request", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Max_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Max_Switch", "OFF")

        # hand timeout
        self._rule_max_without_current.to_Hand()
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_Off")
        tests.helper.oh_item.assert_value("Unittest_Current_Switch", "OFF")

        # manual off
        self._rule_max_without_current.to_Hand()
        tests.helper.oh_item.item_state_change_event("Unittest_Current_Manual", "ON")
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Manual")
        tests.helper.oh_item.assert_value("Unittest_Current_Switch", "OFF")

    def test_to_hand_transitions(self) -> None:
        """Test to Hand transitions."""
        for state in ["Auto_On", "Auto_WaitCurrent", "Auto_Off"]:
            with self.subTest(state=state):
                eval(f"self._rule_with_current.to_{state}()")  # noqa: S307
                tests.helper.oh_item.item_state_change_event("Unittest_Current_Switch", "OFF")
                tests.helper.oh_item.item_state_change_event("Unittest_Current_Switch", "ON")
                tests.helper.oh_item.assert_value("Unittest_Current_State", "Hand")

    def test_manual_transitions(self) -> None:
        """Test Manual transitions."""
        # manual off | on_off_conditions not met
        self._rule_with_current.to_Manual()
        tests.helper.oh_item.item_state_change_event("Unittest_Current_Manual", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_Off")

        # manual off | on_off_conditions met
        self._rule_with_current.to_Manual()
        tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", PresenceState.PRESENCE.value)
        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", SleepState.AWAKE.value)
        tests.helper.oh_item.item_state_change_event("Unittest_Current_Manual", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_On")

    def test_wait_current_extended_transitions(self) -> None:
        """Test WaitCurrentExtended transitions."""
        # on conditions met
        self._rule_with_current.to_Auto_WaitCurrentExtended()
        self._rule_with_current.on_conditions_met()
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_On")

        # current above threshold
        self._rule_with_current.to_Auto_WaitCurrentExtended()
        tests.helper.oh_item.send_command("Unittest_Current", 2)
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_WaitCurrent")

        # extended timeout
        self._rule_with_current.to_Auto_WaitCurrentExtended()
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_Off")

    def test_current_switch_off(self) -> None:
        """Test current switch off."""
        tests.helper.oh_item.set_state("Unittest_Current_Switch", "ON")
        self._rule_with_current.to_Auto_WaitCurrent()
        tests.helper.oh_item.item_state_change_event("Unittest_Current", 2)
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_WaitCurrent")
        tests.helper.oh_item.assert_value("Unittest_Current_Switch", "ON")

        tests.helper.oh_item.item_state_change_event("Unittest_Current", 0.09)
        tests.helper.oh_item.assert_value("Unittest_Current_State", "Auto_WaitCurrentExtended")
        tests.helper.oh_item.assert_value("Unittest_Current_Switch", "ON")



"""Config models for logic rules."""

import HABApp
import pydantic
import typing_extensions

import habapp_rules.core.pydantic_base


class BinaryLogicItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for binary logic."""

    inputs: list[HABApp.openhab.items.SwitchItem | HABApp.openhab.items.ContactItem] = pydantic.Field(..., description="List of input items (must be either Switch or Contact and all have to match to output_item)")
    output: HABApp.openhab.items.SwitchItem | HABApp.openhab.items.ContactItem = pydantic.Field(..., description="Output item")

    @pydantic.model_validator(mode="after")
    def validate_items(self) -> typing_extensions.Self:
        """Validate if all items are of the same type.

        Returns:
            validated model

        Raises:
            TypeError: if not all items are of the same type
        """
        for item in self.inputs:
            if not isinstance(item, type(self.output)):
                msg = f"Item '{item.name}' must have the same type like the output item. Expected: {type(self.output)} | actual : {type(item)}"
                raise TypeError(msg)
        return self


class BinaryLogicConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for binary logic."""

    items: BinaryLogicItems = pydantic.Field(..., description="Items for binary logic")
    parameter: None = None


class NumericLogicItems(BinaryLogicItems):
    """Items for numeric logic."""

    inputs: list[HABApp.openhab.items.NumberItem | HABApp.openhab.items.DimmerItem] = pydantic.Field(..., description="List of input items (must be either Number or Dimmer and all have to match to output_item)")
    output: HABApp.openhab.items.NumberItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="Output item")


class NumericLogicParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for numeric logic."""

    ignore_old_values_time: int | None = pydantic.Field(None, description="ignores values which are older than the given time in seconds. If None, all values will be taken")


class NumericLogicConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for numeric logic."""

    items: NumericLogicItems = pydantic.Field(..., description="Items for numeric logic")
    parameter: NumericLogicParameter = pydantic.Field(NumericLogicParameter(), description="Parameter for numeric logic")


class InvertValueItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for invert value."""

    input: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Input item")
    output: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Output item")


class InvertValueParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for invert value."""

    only_positive: bool = pydantic.Field(default=False, description="if true, only positive values will be set to output item")
    only_negative: bool = pydantic.Field(default=False, description="if true, only negative values will be set to output item")


class InvertValueConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for invert value."""

    items: InvertValueItems = pydantic.Field(..., description="Items for invert value")
    parameter: InvertValueParameter = pydantic.Field(InvertValueParameter(), description="Parameter for invert value")



"""Rules to manage lights."""

from __future__ import annotations

import abc
import copy
import logging
import math
import threading
import time
import typing

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.items
import HABApp.util

import habapp_rules.actors.config.light
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system

if typing.TYPE_CHECKING:
    from collections.abc import Callable  # pragma: no cover

LOGGER = logging.getLogger(__name__)

DIMMER_VALUE_TOLERANCE = 5
MINUTE_IN_SEC = 60


class _LightBase(habapp_rules.core.state_machine_rule.StateMachineRule, abc.ABC):
    """Base class for lights."""

    states: typing.ClassVar = [
        {"name": "manual"},
        {
            "name": "auto",
            "initial": "init",
            "children": [
                {"name": "init"},
                {"name": "on", "timeout": 10, "on_timeout": "auto_on_timeout"},
                {"name": "preoff", "timeout": 4, "on_timeout": "preoff_timeout"},
                {"name": "off"},
                {"name": "leaving", "timeout": 5, "on_timeout": "leaving_timeout"},
                {"name": "presleep", "timeout": 5, "on_timeout": "presleep_timeout"},
                {"name": "restoreState"},
            ],
        },
    ]

    trans: typing.ClassVar = [
        {"trigger": "manual_on", "source": "auto", "dest": "manual"},
        {"trigger": "manual_off", "source": "manual", "dest": "auto"},
        {"trigger": "hand_on", "source": ["auto_off", "auto_preoff"], "dest": "auto_on"},
        {"trigger": "hand_off", "source": ["auto_on", "auto_leaving", "auto_presleep"], "dest": "auto_off"},
        {"trigger": "hand_off", "source": "auto_preoff", "dest": "auto_on"},
        {"trigger": "auto_on_timeout", "source": "auto_on", "dest": "auto_preoff", "conditions": "_pre_off_configured"},
        {"trigger": "auto_on_timeout", "source": "auto_on", "dest": "auto_off", "unless": "_pre_off_configured"},
        {"trigger": "preoff_timeout", "source": "auto_preoff", "dest": "auto_off"},
        {"trigger": "leaving_started", "source": ["auto_on", "auto_off", "auto_preoff"], "dest": "auto_leaving", "conditions": "_leaving_configured"},
        {"trigger": "leaving_aborted", "source": "auto_leaving", "dest": "auto_restoreState"},
        {"trigger": "leaving_timeout", "source": "auto_leaving", "dest": "auto_off"},
        {"trigger": "sleep_started", "source": ["auto_on", "auto_off", "auto_preoff"], "dest": "auto_presleep", "conditions": "_pre_sleep_configured"},
        {"trigger": "sleep_aborted", "source": "auto_presleep", "dest": "auto_restoreState"},
        {"trigger": "presleep_timeout", "source": "auto_presleep", "dest": "auto_off"},
    ]
    _state_observer: habapp_rules.actors.state_observer.StateObserverSwitch | habapp_rules.actors.state_observer.StateObserverDimmer

    def __init__(self, config: habapp_rules.actors.config.light.LightConfig) -> None:
        """Init of basic light object.

        Args:
            config: light config
        """
        self._config = config

        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.light.name)

        # init state machine
        self._previous_state = None
        self._restore_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")

        self._brightness_before = -1
        self._timeout_on = 0
        self._timeout_pre_off = 0
        self._timeout_pre_sleep = 0
        self._timeout_leaving = 0
        self.__time_sleep_start = 0
        self._set_timeouts()
        self._set_initial_state()

        # callbacks
        self._config.items.manual.listen_event(self._cb_manu, HABApp.openhab.events.ItemStateUpdatedEventFilter())
        if self._config.items.sleeping_state is not None:
            self._config.items.sleeping_state.listen_event(self._cb_sleeping, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.presence_state is not None:
            self._config.items.presence_state.listen_event(self._cb_presence, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._config.items.day.listen_event(self._cb_day, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._update_openhab_state()
        self._instance_logger.debug(super().get_initial_log_message())

    def _get_initial_state(self, default_value: str = "") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
             OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._config.items.manual.is_on():
            return "manual"
        if self._config.items.light.is_on():
            if (
                self._config.items.presence_state is not None
                and self._config.items.presence_state.value == habapp_rules.system.PresenceState.PRESENCE.value
                and getattr(self._config.items.sleeping_state, "value", "awake") in {habapp_rules.system.SleepState.AWAKE.value, habapp_rules.system.SleepState.POST_SLEEPING.value, habapp_rules.system.SleepState.LOCKED.value}
            ):
                return "auto_on"
            if (
                self._pre_sleep_configured()
                and self._config.items.presence_state is not None
                and self._config.items.presence_state.value in {habapp_rules.system.PresenceState.PRESENCE.value, habapp_rules.system.PresenceState.LEAVING.value}
                and getattr(self._config.items.sleeping_state, "value", "") in {habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value}
            ):
                return "auto_presleep"
            if self._leaving_configured():
                return "auto_leaving"
            return "auto_on"
        return "auto_off"

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This should method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            self._set_light_state()
            self._previous_state = self.state

    def _pre_off_configured(self) -> bool:
        """Check whether pre-off is configured for the current day/night/sleep state.

        Returns:
            True if pre-off is configured
        """
        return bool(self._timeout_pre_off)

    def _leaving_configured(self) -> bool:
        """Check whether leaving is configured for the current day/night/sleep state.

        Returns:
            True if leaving is configured
        """
        if self._config.parameter.leaving_only_if_on and self._config.items.light.is_off():
            return False

        return bool(self._timeout_leaving)

    def _pre_sleep_configured(self) -> bool:
        """Check whether pre-sleep is configured for the current day/night state.

        Returns:
            True if pre-sleep is configured
        """
        if self._config.items.sleeping_state is None:
            return False

        pre_sleep_prevent = False
        if self._config.items.pre_sleep_prevent is not None:
            pre_sleep_prevent = self._config.items.pre_sleep_prevent.is_on()
        elif self._config.parameter.pre_sleep_prevent is not None:
            try:
                pre_sleep_prevent = self._config.parameter.pre_sleep_prevent()
            except Exception:
                self._instance_logger.exception("Could not execute pre_sleep_prevent function. pre_sleep_prevent will be set to False.")
                pre_sleep_prevent = False

        return bool(self._timeout_pre_sleep) and not pre_sleep_prevent

    def on_enter_auto_restoreState(self) -> None:  # noqa: N802
        """On enter of state auto_restoreState."""
        self._restore_state = "auto_off" if self._restore_state == "auto_preoff" else self._restore_state

        if self._restore_state:
            self._set_state(self._restore_state)

    def _was_on_before(self) -> bool:
        """Check whether the dimmer was on before.

        Returns:
            True if the dimmer was on before, else False
        """
        return bool(self._brightness_before)

    def _set_timeouts(self) -> None:
        """Set timeouts depending on the current day/night/sleep state."""
        if self._get_sleeping_activ():
            self._timeout_on = self._config.parameter.on.sleeping.timeout
            self._timeout_pre_off = getattr(self._config.parameter.pre_off.sleeping if self._config.parameter.pre_off else 0, "timeout", 0)
            self._timeout_leaving = getattr(self._config.parameter.leaving.sleeping if self._config.parameter.leaving else 0, "timeout", 0)
            self._timeout_pre_sleep = 0

        elif self._config.items.day.is_on():
            self._timeout_on = self._config.parameter.on.day.timeout
            self._timeout_pre_off = getattr(self._config.parameter.pre_off.day if self._config.parameter.pre_off else 0, "timeout", 0)
            self._timeout_leaving = getattr(self._config.parameter.leaving.day if self._config.parameter.leaving else 0, "timeout", 0)
            self._timeout_pre_sleep = getattr(self._config.parameter.pre_sleep.day if self._config.parameter.pre_sleep else 0, "timeout", 0)
        else:
            self._timeout_on = self._config.parameter.on.night.timeout
            self._timeout_pre_off = getattr(self._config.parameter.pre_off.night if self._config.parameter.pre_off else 0, "timeout", 0)
            self._timeout_leaving = getattr(self._config.parameter.leaving.night if self._config.parameter.leaving else 0, "timeout", 0)
            self._timeout_pre_sleep = getattr(self._config.parameter.pre_sleep.night if self._config.parameter.pre_sleep else 0, "timeout", 0)

        self.state_machine.states["auto"].states["on"].timeout = self._timeout_on
        self.state_machine.states["auto"].states["preoff"].timeout = self._timeout_pre_off
        self.state_machine.states["auto"].states["leaving"].timeout = self._timeout_leaving
        self.state_machine.states["auto"].states["presleep"].timeout = self._timeout_pre_sleep

    @abc.abstractmethod
    def _set_light_state(self) -> None:
        """Set brightness to light."""

    def _get_target_brightness(self) -> bool | float | None:  # noqa: C901, PLR0912
        """Get configured brightness for the current day/night/sleep state.

        Returns:
             brightness value
        """
        sleeping_active = self._get_sleeping_activ(include_pre_sleep=True)

        if self.state == "auto_on":
            if self._previous_state == "manual":
                return None
            if self._previous_state in {"auto_preoff", "auto_leaving", "auto_presleep"}:
                return self._brightness_before

            # starting from here: previous state == auto_off
            if isinstance(man_value := self._state_observer.last_manual_event.value, int | float) and 0 < man_value < 100:  # noqa: PLR2004
                return None
            if self._state_observer.last_manual_event.value == "INCREASE":
                return None

            if sleeping_active:
                brightness_from_config = self._config.parameter.on.sleeping.brightness
            elif self._config.items.day.is_on():
                brightness_from_config = self._config.parameter.on.day.brightness
            else:
                brightness_from_config = self._config.parameter.on.night.brightness

            if brightness_from_config is True and self._state_observer.last_manual_event.value == "ON":
                return None

            return brightness_from_config

        if self.state == "auto_preoff":
            self._brightness_before = self._state_observer.value

            if sleeping_active:
                brightness_from_config = getattr(self._config.parameter.pre_off.sleeping if self._config.parameter.pre_off else None, "brightness", None)
            elif self._config.items.day.is_on():
                brightness_from_config = getattr(self._config.parameter.pre_off.day if self._config.parameter.pre_off else None, "brightness", None)
            else:
                brightness_from_config = getattr(self._config.parameter.pre_off.night if self._config.parameter.pre_off else None, "brightness", None)

            if brightness_from_config is None:
                return None

            if isinstance(self._state_observer.value, float | int) and brightness_from_config > self._state_observer.value:
                return math.ceil(self._state_observer.value / 2)
            return brightness_from_config

        if self.state == "auto_off":
            if self._previous_state == "manual":
                return None
            return False

        if self.state == "auto_presleep":
            if self._config.items.day.is_on():
                return getattr(self._config.parameter.pre_sleep.day if self._config.parameter.pre_sleep else None, "brightness", None)
            return getattr(self._config.parameter.pre_sleep.night if self._config.parameter.pre_sleep else None, "brightness", None)

        if self.state == "auto_leaving":
            if sleeping_active:
                return getattr(self._config.parameter.leaving.sleeping if self._config.parameter.leaving else None, "brightness", None)
            if self._config.items.day.is_on():
                return getattr(self._config.parameter.leaving.day if self._config.parameter.leaving else None, "brightness", None)
            return getattr(self._config.parameter.leaving.night if self._config.parameter.leaving else None, "brightness", None)

        return None

    def on_enter_auto_init(self) -> None:
        """Callback, which is called on enter of init state."""
        if self._config.items.light.is_on():
            self.to_auto_on()
        else:
            self.to_auto_off()

    def _get_sleeping_activ(self, include_pre_sleep: bool = False) -> bool:
        """Get if sleeping is active.

        Args:
            include_pre_sleep: if true, also pre sleep will be handled as sleeping

        Returns:
            true if sleeping active
        """
        sleep_states = [habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value] if include_pre_sleep else [habapp_rules.system.SleepState.SLEEPING.value]
        return getattr(self._config.items.sleeping_state, "value", "") in sleep_states

    def _cb_hand_on(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:  # noqa: ARG002
        """Callback, which is triggered by the state observer if a manual ON command was detected.

        Args:
            event: original trigger event
        """
        self._instance_logger.debug("Hand 'ON' detected")
        self.hand_on()

    def _cb_hand_off(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:  # noqa: ARG002
        """Callback, which is triggered by the state observer if a manual OFF command was detected.

        Args:
            event: original trigger event
        """
        self._instance_logger.debug("Hand 'OFF' detected")
        self.hand_off()

    def _cb_manu(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
        """Callback, which is triggered if the manual switch has a state event.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.manual_on()
        else:
            self.manual_off()

    def _cb_day(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        """Callback, which is triggered if the day/night switch has a state change event.

        Args:
            event: trigger event
        """
        self._set_timeouts()

    def _cb_presence(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the presence state has a state change event.

        Args:
            event: trigger event
        """
        self._set_timeouts()
        if event.value == habapp_rules.system.PresenceState.LEAVING.value:
            self._brightness_before = self._state_observer.value
            self._restore_state = self._previous_state
            self.leaving_started()
        elif event.value == habapp_rules.system.PresenceState.PRESENCE.value and self.state == "auto_leaving":
            self.leaving_aborted()

    def _cb_sleeping(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the sleep state has a state change event.

        Args:
            event: trigger event
        """
        self._set_timeouts()
        if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
            self._brightness_before = self._state_observer.value
            self._restore_state = self._previous_state
            self.__time_sleep_start = time.time()
            self.sleep_started()
        elif event.value == habapp_rules.system.SleepState.AWAKE.value and time.time() - self.__time_sleep_start <= MINUTE_IN_SEC:
            self.sleep_aborted()


class LightSwitch(_LightBase):
    """Rules class to manage basic light states.

    # KNX-things:
    Thing device T00_99_OpenHab_DimmerObserver "KNX OpenHAB dimmer observer"{
            Type switch             : light             "Light"             [ switch="1/1/10+1/1/13" ]
    }

    # Items:
    Switch    I01_01_Light              "Light"             {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light"}
    Switch    I00_00_Light_manual       "Light manual"

    # Config:
    config = habapp_rules.actors.config.light.LightConfig(
            items = habapp_rules.actors.config.light.LightItems(
                    light="I01_01_Light",
                    manual="I00_00_Light_manual",
                    presence_state="I999_00_Presence_state",
                    sleeping_state="I999_00_Sleeping_state",
                    day="I999_00_Day"
            )
    )

    # Rule init:
    habapp_rules.actors.light.LightSwitch(config)
    """

    def __init__(self, config: habapp_rules.actors.config.light.LightConfig) -> None:
        """Init of basic light object.

        Args:
            config: light config

        Raises:
            TypeError: if type of light_item is not supported
        """
        if not isinstance(config.items.light, HABApp.openhab.items.switch_item.SwitchItem):
            msg = f"type: {type(config.items.light)} is not supported!"
            raise TypeError(msg)

        self._state_observer = habapp_rules.actors.state_observer.StateObserverSwitch(config.items.light.name, self._cb_hand_on, self._cb_hand_off)

        _LightBase.__init__(self, config)

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states."""
        _LightBase._update_openhab_state(self)  # noqa: SLF001

        if self.state == "auto_preoff":
            timeout = self.state_machine.get_state(self.state).timeout

            warn_thread_1 = threading.Thread(target=self.__trigger_warning, args=("auto_preoff", 0, 1), daemon=True)
            warn_thread_1.start()

            if timeout > MINUTE_IN_SEC:
                # add additional warning for long timeouts
                warn_thread_2 = threading.Thread(target=self.__trigger_warning, args=("auto_preoff", timeout / 2, 2), daemon=True)
                warn_thread_2.start()

    def __trigger_warning(self, state_name: str, wait_time: float, switch_off_amount: int) -> None:
        """Trigger light switch off warning.

        Args:
            state_name: name of state where the warning should be triggered. If different no command will be sent
            wait_time: time between start of the thread and switch off / on
            switch_off_amount: number of switch off
        """
        if wait_time:
            time.sleep(wait_time)

        for idx in range(switch_off_amount):
            if self.state != state_name:
                break
            self._state_observer.send_command("OFF")
            time.sleep(0.2)
            if self.state != state_name:
                break
            self._state_observer.send_command("ON")
            if idx + 1 < switch_off_amount:
                time.sleep(0.5)

    def _set_light_state(self) -> None:
        """Set brightness to light."""
        target_value = self._get_target_brightness()
        if target_value is None or self._previous_state is None:
            # don't change value if target_value is None or _set_light_state will be called during init (_previous_state == None)
            return

        target_value = "ON" if target_value else "OFF"
        self._instance_logger.debug(f"set brightness {target_value}")
        self._state_observer.send_command(target_value)


class LightDimmer(_LightBase):
    """Rules class to manage basic light states.

    # KNX-things:
    Thing device T00_99_OpenHab_DimmerObserver "KNX OpenHAB dimmer observer"{
            Type dimmer             : light             "Light"             [ switch="1/1/10", position="1/1/13+<1/1/15" ]
            Type dimmer-control     : light_ctr         "Light control"     [ increaseDecrease="1/1/12"]
            Type dimmer             : light_group       "Light Group"       [ switch="1/1/240", position="1/1/243"]
    }

    # Items:
    Dimmer    I01_01_Light              "Light"             {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light"}
    Dimmer    I01_01_Light_ctr          "Light ctr"         {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_ctr"}
    Dimmer    I01_01_Light_group        "Light Group"       {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_group"}
    Switch    I00_00_Light_manual       "Light manual"

    # Config:
    config = habapp_rules.actors.config.light.LightConfig(
            items=habapp_rules.actors.config.light.LightItems(
                    light="I01_01_Light",
                    light_control=["I01_01_Light_ctr"],
                    manual="I00_00_Light_manual",
                    presence_state="I999_00_Presence_state",
                    sleeping_state="I999_00_Sleeping_state",
                    day="I999_00_Day"
            )
    )

    # Rule init:
    habapp_rules.actors.light.LightDimmer(config)
    """

    trans = copy.deepcopy(_LightBase.trans)
    trans.append({"trigger": "hand_changed", "source": "auto_on", "dest": "auto_on"})

    def __init__(self, config: habapp_rules.actors.config.light.LightConfig) -> None:
        """Init of basic light object.

        Args:
            config: light config

        Raises:
            TypeError: if type of light_item is not supported
        """
        if not isinstance(config.items.light, HABApp.openhab.items.dimmer_item.DimmerItem):
            msg = f"type: {type(config.items.light)} is not supported!"
            raise TypeError(msg)

        control_names = [item.name for item in config.items.light_control]
        group_names = [item.name for item in config.items.light_groups]
        self._state_observer = habapp_rules.actors.state_observer.StateObserverDimmer(config.items.light.name, self._cb_hand_on, self._cb_hand_off, self._cb_hand_changed, control_names=control_names, group_names=group_names)

        _LightBase.__init__(self, config)

    def _set_light_state(self) -> None:
        """Set brightness to light."""
        target_value = self._get_target_brightness()
        if target_value is None or self._previous_state is None:
            # don't change value if target_value is None or _set_light_state will be called during init (_previous_state == None)
            return

        if isinstance(target_value, bool):
            target_value = "ON" if target_value else "OFF"
        self._instance_logger.debug(f"set brightness {target_value}")
        self._state_observer.send_command(target_value)

    def _cb_hand_changed(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent | HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered by the state observer if a manual OFF command was detected.

        Args:
            event: original trigger event
        """
        if isinstance(event, HABApp.openhab.events.ItemStateChangedEvent) and abs(event.value - event.old_value) > DIMMER_VALUE_TOLERANCE:
            self.hand_changed()


class _LightExtendedMixin:
    """Mixin class for adding door and motion functionality."""

    states: dict
    trans: list
    state: str
    _config: habapp_rules.actors.config.light.LightConfig
    state_machine: habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout
    _get_sleeping_activ: Callable[[bool | None], bool]

    def __init__(self, config: habapp_rules.actors.config.light.LightConfig) -> None:
        """Init mixin class.

        Args:
            config: light config
        """
        self.states = _LightExtendedMixin._add_additional_states(self.states)
        self.trans = _LightExtendedMixin._add_additional_transitions(self.trans)

        self._timeout_motion = 0
        self._timeout_door = 0

        self._hand_off_lock_time = config.parameter.hand_off_lock_time
        self._hand_off_timestamp = 0

    @staticmethod
    def _add_additional_states(states_dict: dict) -> dict:
        """Add additional states for door and motion.

        Args:
            states_dict: current state dictionary

        Returns:
            current + new states
        """
        states_dict = copy.deepcopy(states_dict)
        states_dict[1]["children"].append({"name": "door", "timeout": 999, "on_timeout": "door_timeout"})
        states_dict[1]["children"].append({"name": "motion", "timeout": 999, "on_timeout": "motion_timeout"})
        return states_dict

    @staticmethod
    def _add_additional_transitions(transitions_list: list[dict]) -> list[dict]:
        """Add additional transitions for door and motion.

        Args:
            transitions_list: current transitions

        Returns:
            current + new transitions
        """
        transitions_list = copy.deepcopy(transitions_list)

        transitions_list.append({"trigger": "motion_on", "source": "auto_door", "dest": "auto_motion", "conditions": "_motion_configured"})
        transitions_list.append({"trigger": "motion_on", "source": "auto_off", "dest": "auto_motion", "conditions": ["_motion_configured", "_motion_door_allowed"]})
        transitions_list.append({"trigger": "motion_on", "source": "auto_preoff", "dest": "auto_motion", "conditions": "_motion_configured"})
        transitions_list.append({"trigger": "motion_off", "source": "auto_motion", "dest": "auto_preoff", "conditions": "_pre_off_configured"})
        transitions_list.append({"trigger": "motion_off", "source": "auto_motion", "dest": "auto_off", "unless": "_pre_off_configured"})
        transitions_list.append({"trigger": "motion_timeout", "source": "auto_motion", "dest": "auto_preoff", "conditions": "_pre_off_configured", "before": "_log_motion_timeout_warning"})
        transitions_list.append({"trigger": "motion_timeout", "source": "auto_motion", "dest": "auto_off", "unless": "_pre_off_configured", "before": "_log_motion_timeout_warning"})
        transitions_list.append({"trigger": "hand_off", "source": "auto_motion", "dest": "auto_off"})

        transitions_list.append({"trigger": "door_opened", "source": ["auto_off", "auto_preoff", "auto_door"], "dest": "auto_door", "conditions": ["_door_configured", "_motion_door_allowed"]})
        transitions_list.append({"trigger": "door_timeout", "source": "auto_door", "dest": "auto_preoff", "conditions": "_pre_off_configured"})
        transitions_list.append({"trigger": "door_timeout", "source": "auto_door", "dest": "auto_off", "unless": "_pre_off_configured"})
        transitions_list.append({"trigger": "door_closed", "source": "auto_leaving", "dest": "auto_off", "conditions": "_door_off_leaving_configured"})
        transitions_list.append({"trigger": "hand_off", "source": "auto_door", "dest": "auto_off"})

        transitions_list.append({"trigger": "leaving_started", "source": ["auto_motion", "auto_door"], "dest": "auto_leaving", "conditions": "_leaving_configured"})
        transitions_list.append({"trigger": "sleep_started", "source": ["auto_motion", "auto_door"], "dest": "auto_presleep", "conditions": "_pre_sleep_configured"})

        return transitions_list

    def add_additional_callbacks(self) -> None:
        """Add additional callbacks for motion and door items."""
        if self._config.items.motion is not None:
            self._config.items.motion.listen_event(self._cb_motion, HABApp.openhab.events.ItemStateChangedEventFilter())
        for item_door in self._config.items.doors:
            item_door.listen_event(self._cb_door, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _get_initial_state(self, default_value: str = "") -> str:
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        initial_state = _LightBase._get_initial_state(self, default_value)  # noqa: SLF001

        if initial_state == "auto_on" and self._config.items.motion is not None and self._config.items.motion.is_on() and self._motion_configured():
            initial_state = "auto_motion"
        return initial_state

    def _set_timeouts(self) -> None:
        """Set timeouts depending on the current day/night/sleep state."""
        _LightBase._set_timeouts(self)  # noqa: SLF001

        # set timeouts of additional states
        if self._get_sleeping_activ():
            self._timeout_motion = getattr(self._config.parameter.motion.sleeping if self._config.parameter.motion else 0, "timeout", 0)
            self._timeout_door = getattr(self._config.parameter.door.sleeping if self._config.parameter.door else 0, "timeout", 0)

        elif self._config.items.day.is_on():
            self._timeout_motion = getattr(self._config.parameter.motion.day if self._config.parameter.motion else 0, "timeout", 0)
            self._timeout_door = getattr(self._config.parameter.door.day if self._config.parameter.door else 0, "timeout", 0)
        else:
            self._timeout_motion = getattr(self._config.parameter.motion.night if self._config.parameter.motion else 0, "timeout", 0)
            self._timeout_door = getattr(self._config.parameter.door.night if self._config.parameter.door else 0, "timeout", 0)

        self.state_machine.states["auto"].states["motion"].timeout = self._timeout_motion
        self.state_machine.states["auto"].states["door"].timeout = self._timeout_door

    def _get_target_brightness(self) -> bool | float | None:
        """Get configured brightness for the current day/night/sleep state. Must be called before _get_target_brightness of base class.

        Returns:
            configured brightness value

        Raises:
            habapp_rules.core.exceptions.HabAppRulesError: if current state is not supported
        """
        if self.state == "auto_motion":
            if self._get_sleeping_activ(include_pre_sleep=True):
                return getattr(self._config.parameter.motion.sleeping if self._config.parameter.motion else None, "brightness", None)
            if self._config.items.day.is_on():
                return getattr(self._config.parameter.motion.day if self._config.parameter.motion else None, "brightness", None)
            return getattr(self._config.parameter.motion.night if self._config.parameter.motion else None, "brightness", None)

        if self.state == "auto_door":
            if self._get_sleeping_activ(include_pre_sleep=True):
                return getattr(self._config.parameter.door.sleeping if self._config.parameter.door else None, "brightness", None)
            if self._config.items.day.is_on():
                return getattr(self._config.parameter.door.day if self._config.parameter.door else None, "brightness", None)
            return getattr(self._config.parameter.door.night if self._config.parameter.door else None, "brightness", None)

        return _LightBase._get_target_brightness(self)  # noqa: SLF001

    def _door_configured(self) -> bool:
        """Check whether door functionality is configured for the current day/night state.

        Returns:
            True if door functionality is configured
        """
        if not self._config.items.doors:
            return False
        return bool(self._timeout_door)

    def _door_off_leaving_configured(self) -> bool:
        """Check whether door-off functionality is configured for the current day/night state.

        Returns:
            True if door-off is configured
        """
        return self._config.parameter.off_at_door_closed_during_leaving

    def _motion_configured(self) -> bool:
        """Check whether motion functionality is configured for the current day/night state.

        Returns:
            True if motion functionality is configured
        """
        if self._config.items.motion is None:
            return False
        return bool(self._timeout_motion)

    def _motion_door_allowed(self) -> bool:
        """Check if transition to motion and door state is allowed.

        Returns:
            True if transition is allowed
        """
        return time.time() - self._hand_off_timestamp > self._hand_off_lock_time

    def _cb_hand_off(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Callback, which is triggered by the state observer if a manual OFF command was detected.

        Args:
            event: original trigger event
        """
        self._hand_off_timestamp = time.time()
        _LightBase._cb_hand_off(self, event)  # noqa: SLF001

    def _cb_motion(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the motion state changed.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.motion_on()
        else:
            self.motion_off()

    def _cb_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if a door state changed.

        Args:
            event: trigger event
        """
        if event.value == "OPEN":
            # every open of a single door calls door_opened()
            self.door_opened()

        if event.value == "CLOSED" and all(door.is_closed() for door in self._config.items.doors):
            # only if all doors are closed door_closed() is called
            self.door_closed()

    def _log_motion_timeout_warning(self) -> None:
        """Log warning if motion state was left because of timeout."""
        self._instance_logger.warning("Timeout of motion was triggered, before motion stopped. Thing about to increase motion timeout!")


class LightSwitchExtended(_LightExtendedMixin, LightSwitch):
    """Extended Light.

    Example config is given at Light base class.
    With this class additionally motion or door items can be given.
    """

    def __init__(self, config: habapp_rules.actors.config.light.LightConfig) -> None:
        """Init of extended light object.

        Args:
            config: light config
        """
        _LightExtendedMixin.__init__(self, config)
        LightSwitch.__init__(self, config)

        _LightExtendedMixin.add_additional_callbacks(self)


class LightDimmerExtended(_LightExtendedMixin, LightDimmer):
    """Extended Light.

    Example config is given at Light base class.
    With this class additionally motion or door items can be given.
    """

    def __init__(self, config: habapp_rules.actors.config.light.LightConfig) -> None:
        """Init of extended light object.

        Args:
            config: light config
        """
        _LightExtendedMixin.__init__(self, config)
        LightDimmer.__init__(self, config)

        _LightExtendedMixin.add_additional_callbacks(self)



"""Config models for light rules."""

import collections.abc
import logging

import HABApp.openhab.items
import pydantic
import typing_extensions

import habapp_rules.core.exceptions
import habapp_rules.core.pydantic_base

LOGGER = logging.getLogger(__name__)


class LightItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for all light rules."""

    light: HABApp.openhab.items.DimmerItem | HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which controls the light")
    light_control: list[HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="control items to improve manual detection")
    light_groups: list[HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="group items which can additionally set the light state. This can be used to improve the manual detection")
    manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to switch to manual mode and disable the automatic functions")
    presence_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="presence state set via habapp_rules.presence.Presence")
    day: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which is ON at day and OFF at night")
    sleeping_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleeping state set via habapp_rules.system.sleep.Sleep")
    pre_sleep_prevent: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item to prevent pre-sleep (Can be used for example to prevent the pre sleep light when guests are sleeping)")
    doors: list[HABApp.openhab.items.ContactItem] = pydantic.Field([], description="door items for switching on the light if the door is opening")
    motion: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="motion sensor to enable light if motion is detected")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the current state of the state machine")


class BrightnessTimeout(pydantic.BaseModel):
    """Define brightness and timeout for light states."""

    brightness: int | bool = pydantic.Field(..., description="brightness which should be set. If bool ON will be sent for True and OFF for False")
    timeout: float = pydantic.Field(..., description="Timeout / max time in seconds until switch off")

    def __init__(self, brightness: int | bool, timeout: float) -> None:
        """Initialize BrightnessTimeout without kwargs.

        Args:
            brightness: brightness value
            timeout: timeout value
        """
        super().__init__(brightness=brightness, timeout=timeout)

    @pydantic.model_validator(mode="after")
    def validata_model(self) -> typing_extensions.Self:
        """Validate brightness and timeout.

        Returns:
            self

        Raises:
            AssertionError: if brightness and timeout are not valid
        """
        if self.brightness is False or self.brightness == 0:  # noqa: SIM102
            # Default if the light should be switched off e.g. for leaving / sleeping
            if not self.timeout:
                self.timeout = 0.5

        if not self.timeout:
            msg = f"Brightness and timeout are not valid: brightness = {self.brightness} | timeout = {self.timeout}"
            raise AssertionError(msg)
        return self


class FunctionConfig(pydantic.BaseModel):
    """Define brightness and timeout values for one function."""

    day: BrightnessTimeout | None = pydantic.Field(..., description="config for day. If None the light will not be switched on during the day")
    night: BrightnessTimeout | None = pydantic.Field(..., description="config for night. If None the light will not be switched on during the night")
    sleeping: BrightnessTimeout | None = pydantic.Field(..., description="config for sleeping. If None the light will not be switched on during sleeping")


class LightParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for all light rules.

    For all parameter which have the following type: FunctionConfig | None -> If None this state will be disabled for the rule
    """

    on: FunctionConfig = pydantic.Field(
        FunctionConfig(day=BrightnessTimeout(brightness=True, timeout=14 * 3600), night=BrightnessTimeout(80, 10 * 3600), sleeping=BrightnessTimeout(20, 3 * 3600)), description="values which are used if the light is switched on manually"
    )
    pre_off: FunctionConfig | None = pydantic.Field(FunctionConfig(day=BrightnessTimeout(50, 10), night=BrightnessTimeout(40, 7), sleeping=BrightnessTimeout(10, 7)), description="values which are used if the light changes pre_off state")
    leaving: FunctionConfig | None = pydantic.Field(
        FunctionConfig(day=BrightnessTimeout(brightness=False, timeout=0), night=BrightnessTimeout(brightness=False, timeout=0), sleeping=BrightnessTimeout(brightness=False, timeout=0)),
        description="values which are used if the light changes to leaving state",
    )
    pre_sleep: FunctionConfig | None = pydantic.Field(
        FunctionConfig(day=BrightnessTimeout(brightness=False, timeout=10), night=BrightnessTimeout(brightness=False, timeout=10), sleeping=None), description="values which are used if the light changes to pre_sleep state"
    )
    pre_sleep_prevent: collections.abc.Callable[[], bool] | HABApp.openhab.items.OpenhabItem | None = pydantic.Field(None, description="Enable pre sleep prevent -> disable pre sleep if True")
    motion: FunctionConfig | None = pydantic.Field(None, description="values which are used if the light changes to motion state")
    door: FunctionConfig | None = pydantic.Field(None, description="values which are used if the light is enabled via a door opening")
    off_at_door_closed_during_leaving: bool = pydantic.Field(default=False, description="this can be used to switch lights off, when door is closed in leaving state")
    hand_off_lock_time: int = pydantic.Field(20, description="time in seconds where door / motion switch on is disabled after a manual OFF")
    leaving_only_if_on: bool = pydantic.Field(default=False, description="switch to leaving only if light is on. If False leaving light is always activated")

    @pydantic.field_validator("on", mode="after")
    @classmethod
    def validate_on(cls, value: FunctionConfig) -> FunctionConfig:
        """Validate config for on-state.

        Args:
            value: given value

        Returns:
            validated value

        Raises:
             AssertionError: if on is not valid
        """
        if any(conf is None for conf in [value.day, value.night, value.sleeping]):
            msg = "For function 'on' all brightness / timeout values must be set."
            raise AssertionError(msg)
        return value

    @pydantic.field_validator("pre_sleep", mode="after")
    @classmethod
    def validate_pre_sleep(cls, value: FunctionConfig | None) -> FunctionConfig | None:
        """Validate pre_sleep config.

        Args:
            value: value of pre sleep

        Returns:
            validated value

        Raises:
            AssertionError: if pre_sleep is not valid
        """
        if value is None:
            return value

        if value.sleeping is not None:
            LOGGER.warning("It's not allowed to set brightness / timeout for pre_sleep.sleeping. Set it to None")
            value.sleeping = None

        return value


class LightConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for all light rules."""

    items: LightItems = pydantic.Field(..., description="items for all light rules")
    parameter: LightParameter = pydantic.Field(LightParameter(), description="parameter for all light rules")

    @pydantic.model_validator(mode="after")
    def validate_config(self) -> typing_extensions.Self:
        """Validate config.

        Returns:
            validated config

        Raises:
            AssertionError: if config is not valid
        """
        if self.items.motion is not None and self.parameter.motion is None:
            msg = "item motion is given, but not configured via parameter"
            raise AssertionError(msg)

        if len(self.items.doors) and self.parameter.door is None:
            msg = "item door is given, but not configured via parameter"
            raise AssertionError(msg)

        if self.items.sleeping_state is not None and self.parameter.pre_sleep is None:
            msg = "item sleeping_state is given, but not configured via parameter"
            raise AssertionError(msg)

        if self.items.presence_state is not None and self.parameter.leaving is None:
            msg = "item presence_state is given, but not configured via parameter"
            raise AssertionError(msg)

        if self.items.pre_sleep_prevent is not None and self.parameter.pre_sleep_prevent is not None:
            LOGGER.warning("item pre_sleep_prevent and parameter pre_sleep_prevent are given. The item will be prioritized and the parameter will be ignored!")

        return self



"""Rule to set/unset sleep state."""

import datetime
import logging
import typing

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.items
import HABApp.util

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system.config.sleep

LOGGER = logging.getLogger(__name__)


class Sleep(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Rules class to manage sleep state.

    Example OpenHAB configuration:
    # KNX-things:
    Thing device T00_99_OpenHab_Sleep "KNX OpenHAB Sleep"{
        Type switch             : sleep             "Sleep Request"             [ ga="0/2/30"]
        Type switch-control     : sleep_RM          "Sleep RM"                  [ ga="0/2/31"]

        Type switch             : sleep_lock        "Sleep Lock Request"        [ ga="0/2/32"]
        Type switch-control     : sleep_lock_RM     "Sleep Lock RM"             [ ga="0/2/33"]

        Type string-control     : sleep_text        "Sleep Text"                [ ga="16.000:0/2/34"]
    }

    # Items:
    Switch    I01_02_Sleep              "Sleep"                     <moon>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_RM"}
    Switch    I01_02_Sleep_req          "Sleep request"             <moon>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep"}
    String    I01_02_Sleep_text         "Text for display"                     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_text"}
    Switch    I01_02_Sleep_lock         "Lock"                      <lock>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_lock_RM"}
    Switch    I01_02_Sleep_lock_req     "Lock request"              <lock>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_lock"}
    String    I01_02_Sleep_State        "State"                     <state>

    # Config:
    config = habapp_rules.system.config.sleep.SleepConfig(
            items=habapp_rules.system.config.sleep.SleepItems(
                    sleep="I01_02_Sleep",
                    sleep_req="I01_02_Sleep_req",
                    state="I01_02_Sleep_State",
                    lock="I01_02_Sleep_lock",
                    lock_req="I01_02_Sleep_lock_req",
                    display_text="I01_02_Sleep_text"
            )
    )

    # Rule init:
    habapp_rules.system.sleep.Sleep(config)
    """

    states: typing.ClassVar = [
        {"name": "awake"},
        {"name": "pre_sleeping", "timeout": 3, "on_timeout": "pre_sleeping_timeout"},
        {"name": "sleeping"},
        {"name": "post_sleeping", "timeout": 3, "on_timeout": "post_sleeping_timeout"},
        {"name": "locked"},
    ]

    trans: typing.ClassVar = [
        {"trigger": "start_sleeping", "source": "awake", "dest": "pre_sleeping"},
        {"trigger": "pre_sleeping_timeout", "source": "pre_sleeping", "dest": "sleeping"},
        {"trigger": "end_sleeping", "source": "sleeping", "dest": "post_sleeping"},
        {"trigger": "end_sleeping", "source": "pre_sleeping", "dest": "awake", "unless": "lock_request_active"},
        {"trigger": "end_sleeping", "source": "pre_sleeping", "dest": "locked", "conditions": "lock_request_active"},
        {"trigger": "post_sleeping_timeout", "source": "post_sleeping", "dest": "awake", "unless": "lock_request_active"},
        {"trigger": "post_sleeping_timeout", "source": "post_sleeping", "dest": "locked", "conditions": "lock_request_active"},
        {"trigger": "set_lock", "source": "awake", "dest": "locked"},
        {"trigger": "release_lock", "source": "locked", "dest": "awake"},
    ]

    def __init__(self, config: habapp_rules.system.config.sleep.SleepConfig) -> None:
        """Init of Sleep object.

        Args:
            config: config for sleeping state
        """
        self._config = config
        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.sleep.name)

        # init attributes
        self._sleep_request_active = config.items.sleep_request.is_on()
        self._lock_request_active = config.items.lock_request.is_on() if config.items.lock_request is not None else False

        # init state machine
        self.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()

        self._update_openhab_state()

        # add callbacks
        config.items.sleep_request.listen_event(self._cb_sleep_request, HABApp.openhab.events.ItemStateChangedEventFilter())
        if config.items.lock_request is not None:
            config.items.lock_request.listen_event(self._cb_lock_request, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._instance_logger.debug(super().get_initial_log_message())

    def _get_initial_state(self, default_value: str = "awake") -> str:
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            return correct state if it could be detected, if not return default value
        """
        sleep_req = self._config.items.sleep_request.is_on() if self._config.items.sleep_request.value is not None else None
        lock_req = self._config.items.lock_request.is_on() if self._config.items.lock_request is not None and self._config.items.lock_request.value is not None else None

        if sleep_req:
            return "sleeping"
        if lock_req:
            return "locked"
        if sleep_req is False:
            return "awake"

        return default_value

    @property
    def sleep_request_active(self) -> bool:
        """Check if a sleep request is active.

        Returns:
            return true if lock request is active
        """
        return self._sleep_request_active

    @property
    def lock_request_active(self) -> bool:
        """Check if a lock request is active.

        Returns:
            return true if lock request is active
        """
        return self._lock_request_active

    def _update_openhab_state(self) -> None:
        """Extend _update_openhab state of base class to also update other OpenHAB items."""
        super()._update_openhab_state()

        # update sleep state
        if self.state in {"pre_sleeping", "sleeping"}:
            habapp_rules.core.helper.send_if_different(self._config.items.sleep, "ON")
        else:
            habapp_rules.core.helper.send_if_different(self._config.items.sleep, "OFF")

        # update lock state
        self.__update_lock_state()

        # update display text
        if self._config.items.display_text is not None:
            self._config.items.display_text.oh_send_command(self.__get_display_text())

    def __get_display_text(self) -> str:
        """Get Text for displays.

        Returns:
            display text
        """
        if self.state == "awake":
            return "Schlafen"
        if self.state == "pre_sleeping":
            return "Guten Schlaf"
        if self.state == "sleeping":
            return "Aufstehen"
        if self.state == "post_sleeping":
            return "Guten Morgen"
        if self.state == "locked":
            return "Gesperrt"
        return ""

    def __update_lock_state(self) -> None:
        """Update the return lock state value of OpenHAB item."""
        if self._config.items.lock is not None:
            if self.state in {"pre_sleeping", "post_sleeping", "locked"}:
                habapp_rules.core.helper.send_if_different(self._config.items.lock, "ON")
            else:
                habapp_rules.core.helper.send_if_different(self._config.items.lock, "OFF")

    def _cb_sleep_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if sleep request item changed state.

        Args:
            event: Item state change event of sleep_request item
        """
        if event.value == "ON" and self.state == "awake":
            self._instance_logger.debug("Start sleeping through sleep switch")
            self._sleep_request_active = True
            self.start_sleeping()
        elif event.value == "ON" and self.state == "locked":
            self._sleep_request_active = False
            self._config.items.sleep_request.oh_send_command("OFF")
        elif event.value == "OFF" and self.state in {"sleeping", "pre_sleeping"}:
            self._instance_logger.debug("End sleeping through sleep switch")
            self._sleep_request_active = True
            self.end_sleeping()

    def _cb_lock_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if lock request item changed state.

        Args:
            event: Item state change event of sleep_request item
        """
        self._lock_request_active = event.value == "ON"

        if self.state == "awake" and event.value == "ON":
            self.set_lock()
        elif self.state == "locked" and event.value == "OFF":
            self.release_lock()
        else:
            self.__update_lock_state()


class LinkSleep(HABApp.Rule):
    """Link sleep items depending on current time."""

    def __init__(self, config: habapp_rules.system.config.sleep.LinkSleepConfig) -> None:
        """Init rule.

        Args:
            config: Config for linking sleep rules
        """
        self._config = config
        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self.rule_name)

        config.items.sleep_master.listen_event(self._cb_master, HABApp.openhab.events.ItemStateChangedEventFilter())

        if config.items.link_active_feedback is not None:
            self.run.at(self.run.trigger.time(config.parameter.link_time_start), self._set_link_active_feedback, target_state="ON")
            self.run.at(self.run.trigger.time(config.parameter.link_time_end), self._set_link_active_feedback, target_state="OFF")
            self.run.soon(self._set_link_active_feedback, target_state=self._check_time_in_window())

    def _check_time_in_window(self) -> bool:
        """Check if current time is in the active time window.

        Returns:
            True if current time is in time the active time window
        """
        now = datetime.datetime.now().time()

        if self._config.parameter.link_time_start <= self._config.parameter.link_time_end:
            return self._config.parameter.link_time_start <= now <= self._config.parameter.link_time_end
        # cross midnight
        return self._config.parameter.link_time_start <= now or now <= self._config.parameter.link_time_end

    def _cb_master(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the state of the master item changes.

        Args:
            event: state change event
        """
        if not self._check_time_in_window():
            return

        self._instance_logger.debug(f"Set request of all linked sleep states of {self._config.items.sleep_master.name}")
        for itm in self._config.items.sleep_request_slaves:
            habapp_rules.core.helper.send_if_different(itm, event.value)

    def _set_link_active_feedback(self, target_state: str) -> None:
        """Set feedback for link is active.

        Args:
            target_state: Target state which should be set ["ON" / "OFF"]
        """
        self._config.items.link_active_feedback.oh_send_command(target_state)



import HABApp
import pydantic
from pydantic_extra_types.mac_address import MacAddress

import habapp_rules.core.pydantic_base


class WolItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for WOL rule."""

    trigger_wol: HABApp.openhab.items.switch_item.SwitchItem = pydantic.Field(..., description="item which triggers the WOL")


class WolParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for WOL rule."""

    mac_address: MacAddress = pydantic.Field(..., description="MAC address of the device to wake up")
    friendly_name: str | None = pydantic.Field(None, description="Name which is used for logging")

    @property
    def log_name(self) -> str:
        """Get name for logging.

        Returns:
            Name which is can be used for logging
        """
        return self.friendly_name or self.mac_address


class WolConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for WOL rule."""

    items: WolItems = pydantic.Field(..., description="items for WOL")
    parameter: WolParameter = pydantic.Field(..., description="parameter for WOL")



"""Config models for KNX / MQTT bridge."""

import HABApp.openhab.items
import pydantic
import typing_extensions

import habapp_rules.core.pydantic_base


class KnxMqttItems(habapp_rules.core.pydantic_base.ItemBase):
    """Configuration of items for KNX MQTT bridge."""

    mqtt_dimmer: HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="")
    knx_switch_ctr: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="")
    knx_dimmer_ctr: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="")

    @pydantic.model_validator(mode="after")
    def validate_knx_items(self) -> typing_extensions.Self:
        """Validate KNX items.

        Returns:
                validated model

        Raises:
                ValueError: if knx_switch_ctr and knx_dimmer_ctr are not set
        """
        if self.knx_switch_ctr is None and self.knx_dimmer_ctr is None:
            msg = "knx_switch_ctr or knx_dimmer_ctr must be set"
            raise ValueError(msg)
        return self


class KnxMqttParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Configuration of parameters for KNX MQTT bridge."""

    increase_value: int = pydantic.Field(60, description="")
    decrease_value: int = pydantic.Field(30, description="")


class KnxMqttConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Configuration of KNX MQTT bridge."""

    items: KnxMqttItems = pydantic.Field(..., description="Items for KNX MQTT bridge")
    parameter: KnxMqttParameter = pydantic.Field(KnxMqttParameter(), description="Parameters for KNX MQTT bridge")



"""Config models for filter rules."""

import HABApp.openhab.items
import pydantic
import typing_extensions

import habapp_rules.core.pydantic_base


class ExponentialFilterItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for exponential filter."""

    raw: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Item for raw value")
    filtered: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Item for filtered value")


class ExponentialFilterParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for exponential filter."""

    tau: int = pydantic.Field(..., description="filter time constant in seconds. E.g. step from 0 to 1 | tau = 5 seconds -> after 5 seconds the value will be 0,67")
    instant_increase: bool = pydantic.Field(default=False, description="if set to True, increase of input values will not be filtered")
    instant_decrease: bool = pydantic.Field(default=False, description="if set to True, decrease of input values will not be filtered")

    @pydantic.model_validator(mode="after")
    def validate_instant_parameters(self) -> typing_extensions.Self:
        """Validate instant_increase and instant_decrease.

        Returns:
            validated model

        Raises:
            ValueError: if both parameters are set
        """
        if self.instant_decrease and self.instant_increase:
            msg = "instant_increase and instant_decrease can not be set to True at the same time!"
            raise ValueError(msg)
        return self


class ExponentialFilterConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for exponential filter."""

    items: ExponentialFilterItems = pydantic.Field(..., description="Items for exponential filter")
    parameter: ExponentialFilterParameter = pydantic.Field(..., description="Parameter for exponential filter")



"""Test config models for filter rules."""

import HABApp

import habapp_rules.common.config.filter
import tests.helper.oh_item
import tests.helper.test_case_base


class TestExponentialFilterConfig(tests.helper.test_case_base.TestCaseBase):
    """Test ExponentialFilterConfig."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Raw", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Filtered", 0)

    def test_init(self) -> None:
        """Test __init__."""
        # instant_increase and instant_decrease is not set
        habapp_rules.common.config.filter.ExponentialFilterConfig(
            items=habapp_rules.common.config.filter.ExponentialFilterItems(raw="Unittest_Raw", filtered="Unittest_Filtered"),
            parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(
                tau=42,
            ),
        )

        # instant_increase is set and instant_decrease is not set
        habapp_rules.common.config.filter.ExponentialFilterConfig(
            items=habapp_rules.common.config.filter.ExponentialFilterItems(raw="Unittest_Raw", filtered="Unittest_Filtered"), parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(tau=42, instant_increase=True)
        )

        # instant_increase is not set and instant_decrease is set
        habapp_rules.common.config.filter.ExponentialFilterConfig(
            items=habapp_rules.common.config.filter.ExponentialFilterItems(raw="Unittest_Raw", filtered="Unittest_Filtered"), parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(tau=42, instant_decrease=True)
        )

        # instant_increase and instant_decrease is set
        with self.assertRaises(ValueError):
            habapp_rules.common.config.filter.ExponentialFilterConfig(
                items=habapp_rules.common.config.filter.ExponentialFilterItems(raw="Unittest_Raw", filtered="Unittest_Filtered"),
                parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(tau=42, instant_increase=True, instant_decrease=True),
            )



"""Rules for managing motion sensors."""

import logging
import typing

import HABApp

import habapp_rules.common.hysteresis
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.motion
import habapp_rules.system.sleep

LOGGER = logging.getLogger(__name__)


class Motion(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Class for filtering motion sensors.

    # MQTT-things:
    Thing topic Motion "Motion Sensor"{
        Channels:
        Type switch : motion        "Motion"        [stateTopic="zigbee2mqtt/Motion/occupancy", on="true", off="false"]
        Type number : brightness    "Brightness"    [stateTopic="zigbee2mqtt/Motion/illuminance_lux"]
    }

    # Items:
    Switch    Motion_raw                "Motion raw"                <motion>        {channel="mqtt:topic:broker:Motion:motion"}
    Switch    Motion_filtered           "Motion filtered"           <motion>
    Number    Motion_Brightness         "Brightness"                                {channel="mqtt:topic:broker:Motion:brightness"}
    String    I999_00_Sleeping_state    "Sleeping state"            <state>

    # Config
    config = habapp_rules.sensors.config.motion.MotionConfig(
            items=habapp_rules.sensors.config.motion.MotionItems(
                    motion_raw="Motion_raw",
                    motion_filtered="Motion_filtered",
                    brightness="Motion_Brightness",
                    sleep_state="I999_00_Sleeping_state"
            ),
            parameter=habapp_rules.sensors.config.motion.MotionParameter(
                    brightness_threshold=100,
            ),
    )

    # Rule init:
    habapp_rules.sensors.motion.Motion(config)
    """

    states: typing.ClassVar = [
        {"name": "Locked"},
        {"name": "SleepLocked"},
        {"name": "PostSleepLocked", "timeout": 99, "on_timeout": "timeout_post_sleep_locked"},
        {
            "name": "Unlocked",
            "initial": "Init",
            "children": [
                {"name": "Init"},
                {"name": "Wait"},
                {"name": "Motion"},
                {"name": "MotionExtended", "timeout": 99, "on_timeout": "timeout_motion_extended"},
                {"name": "TooBright"},
            ],
        },
    ]

    trans: typing.ClassVar = [
        # lock
        {"trigger": "lock_on", "source": ["Unlocked", "SleepLocked", "PostSleepLocked"], "dest": "Locked"},
        {"trigger": "lock_off", "source": "Locked", "dest": "Unlocked", "unless": "_sleep_active"},
        {"trigger": "lock_off", "source": "Locked", "dest": "SleepLocked", "conditions": "_sleep_active"},
        # sleep
        {"trigger": "sleep_started", "source": ["Unlocked", "PostSleepLocked"], "dest": "SleepLocked"},
        {"trigger": "sleep_end", "source": "SleepLocked", "dest": "Unlocked", "unless": "_post_sleep_lock_configured"},
        {"trigger": "sleep_end", "source": "SleepLocked", "dest": "PostSleepLocked", "conditions": "_post_sleep_lock_configured"},
        {"trigger": "timeout_post_sleep_locked", "source": "PostSleepLocked", "dest": "Unlocked", "unless": "_raw_motion_active"},
        {"trigger": "motion_off", "source": "PostSleepLocked", "dest": "PostSleepLocked"},
        {"trigger": "motion_on", "source": "PostSleepLocked", "dest": "PostSleepLocked"},
        # motion
        {"trigger": "motion_on", "source": "Unlocked_Wait", "dest": "Unlocked_Motion"},
        {"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_MotionExtended", "conditions": "_motion_extended_configured"},
        {"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_Wait", "unless": "_motion_extended_configured"},
        {"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Wait", "unless": "_brightness_over_threshold"},
        {"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_TooBright", "conditions": "_brightness_over_threshold"},
        {"trigger": "motion_on", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Motion"},
        # brightness
        {"trigger": "brightness_over_threshold", "source": "Unlocked_Wait", "dest": "Unlocked_TooBright"},
        {"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Wait", "unless": "_raw_motion_active"},
        {"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Motion", "conditions": "_raw_motion_active"},
    ]

    def __init__(self, config: habapp_rules.sensors.config.motion.MotionConfig) -> None:
        """Init of motion filter.

        Args:
            config: config for the motion filter

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not valid
        """
        self._config = config
        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.motion_raw.name)

        self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(threshold := self._get_brightness_threshold(), threshold * 0.1 if threshold else 5) if self._config.items.brightness is not None else None

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()

        self.state_machine.get_state("PostSleepLocked").timeout = self._config.parameter.post_sleep_lock_time
        self.state_machine.get_state("Unlocked_MotionExtended").timeout = self._config.parameter.extended_motion_time

        # register callbacks
        self._config.items.motion_raw.listen_event(self._cb_motion_raw, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.brightness is not None:
            self._config.items.brightness.listen_event(self._cb_brightness, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.brightness_threshold is not None:
            self._config.items.brightness_threshold.listen_event(self._cb_threshold_change, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.lock is not None:
            self._config.items.lock.listen_event(self._cb_lock, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.sleep_state is not None:
            self._config.items.sleep_state.listen_event(self._cb_sleep, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._instance_logger.debug(super().get_initial_log_message())

    def _get_initial_state(self, default_value: str = "initial") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._config.items.lock is not None and self._config.items.lock.is_on():
            return "Locked"
        if self._config.items.sleep_state is not None and self._config.items.sleep_state.value == habapp_rules.system.SleepState.SLEEPING.value:
            return "SleepLocked"
        if self._config.items.brightness is not None and self._brightness_over_threshold():
            return "Unlocked_TooBright"
        if self._config.items.motion_raw.is_on():
            return "Unlocked_Motion"
        return "Unlocked_Wait"

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self.__send_filtered_motion()

            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")
            self._previous_state = self.state

    def __send_filtered_motion(self) -> None:
        """Send filtered motion state to OpenHAB item."""
        target_state = "ON" if self.state in {"Unlocked_Motion", "Unlocked_MotionExtended"} else "OFF"
        if target_state != self._config.items.motion_filtered.value:
            self._config.items.motion_filtered.oh_send_command(target_state)

    def _raw_motion_active(self) -> bool:
        """Check if raw motion is active.

        Returns:
            True if active, else False
        """
        return bool(self._config.items.motion_raw)

    def _brightness_over_threshold(self) -> bool:
        """Check if brightness is over threshold.

        Returns:
             True if active, else False
        """
        return self._hysteresis_switch.get_output(self._config.items.brightness.value)

    def _motion_extended_configured(self) -> bool:
        """Check if extended motion is configured.

        Returns:
            True if active, else False
        """
        return self._config.parameter.extended_motion_time > 0

    def _post_sleep_lock_configured(self) -> bool:
        """Check if post sleep lock is configured.

        Returns:
             rue if active, else False
        """
        return self._config.parameter.post_sleep_lock_time > 0

    def _sleep_active(self) -> bool:
        """Check if sleeping is active.

        Returns:
            True if sleeping is active, else False
        """
        return self._config.items.sleep_state.value == habapp_rules.system.SleepState.SLEEPING.value

    def _get_brightness_threshold(self) -> float:
        """Get the current brightness threshold value.

        Returns:
            brightness threshold

        Raises:
            habapp_rules.core.exceptions.HabAppRulesError: if brightness value not given by item or value
        """
        if self._config.parameter.brightness_threshold:
            return self._config.parameter.brightness_threshold
        if self._config.items.brightness_threshold is not None:
            return value if (value := self._config.items.brightness_threshold.value) else float("inf")
        msg = f"Can not get brightness threshold. Brightness value or item is not given. value: {self._config.parameter.brightness_threshold} | item: {self._config.items.brightness_threshold}"
        raise habapp_rules.core.exceptions.HabAppRulesError(msg)

    def on_enter_Unlocked_Init(self) -> None:  # noqa: N802
        """Callback, which is called on enter of Unlocked_Init state."""
        if self._config.items.brightness is not None and self._brightness_over_threshold():
            self.to_Unlocked_TooBright()
        elif self._config.items.motion_raw.is_on():
            self.to_Unlocked_Motion()
        else:
            self.to_Unlocked_Wait()

    def _check_brightness(self, value: float | None = None) -> None:
        """Check if brightness is higher than the threshold and trigger the class methods.

        Args:
            value: Value to check. None if last value should be used
        """
        if self._hysteresis_switch.get_output(value):
            self.brightness_over_threshold()
        else:
            self.brightness_below_threshold()

    def _cb_threshold_change(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the brightness threshold state changed.

        Args:
            event: trigger event
        """
        self._hysteresis_switch.set_threshold_on(event.value)
        self._check_brightness()

    def _cb_motion_raw(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the raw motion state changed.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.motion_on()
        else:
            self.motion_off()

    def _cb_brightness(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the brightness state changed.

        Args:
            event: trigger event
        """
        self._check_brightness(event.value)

    def _cb_lock(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the lock state changed.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.lock_on()
        else:
            self.lock_off()

    def _cb_sleep(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the sleep state changed.

        Args:
            event: trigger event
        """
        if event.value == habapp_rules.system.SleepState.SLEEPING.value:
            self.sleep_started()
        if event.value == habapp_rules.system.SleepState.AWAKE.value:
            self.sleep_end()



"""Module for sending the monthly energy consumption."""

import datetime
import logging
import pathlib
import tempfile

import dateutil.relativedelta
import HABApp
import HABApp.core.internals
import jinja2
import multi_notifier.connectors.connector_mail

import habapp_rules
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.energy.config.monthly_report
import habapp_rules.energy.donut_chart
import habapp_rules.energy.helper

LOGGER = logging.getLogger(__name__)

MONTH_MAPPING = {1: "Januar", 2: "Februar", 3: "März", 4: "April", 5: "Mai", 6: "Juni", 7: "Juli", 8: "August", 9: "September", 10: "Oktober", 11: "November", 12: "Dezember"}


def _get_previous_month_name() -> str:
    """Get name of the previous month.

    if other languages are required, the global dict must be replaced

    Returns:
        name of current month
    """
    today = datetime.datetime.now()
    last_month = today.replace(day=1) - datetime.timedelta(days=1)

    return MONTH_MAPPING[last_month.month]


class MonthlyReport(HABApp.Rule):
    """Rule for sending the monthly energy consumption.

    # Config
    config = habapp_rules.energy.config.monthly_report.MonthlyReportConfig(
            items=habapp_rules.energy.config.monthly_report.MonthlyReportItems(
                    energy_sum="Total Energy"
            ),
            parameter=habapp_rules.energy.config.monthly_report.MonthlyReportParameter(
                    known_energy_shares=[
                            habapp_rules.energy.config.monthly_report.EnergyShare("Dishwasher_Energy", "Dishwasher"),
                            habapp_rules.energy.config.monthly_report.EnergyShare("Light", "Light")
                    ],
                    config_mail=multi_notifier.connectors.connector_mail.MailConfig(
                            user="sender@test.de",
                            password="fancy_password",
                            smtp_host="smtp.test.de",
                            smtp_port=587
                    ),
                    recipients=["test@test.de"],
            )
    )

    # Rule init
    habapp_rules.energy.monthly_report.MonthlyReport("Total_Energy", known_energy_share, "Group_RRD4J", config_mail, "test@test.de")
    """

    def __init__(self, config: habapp_rules.energy.config.monthly_report.MonthlyReportConfig) -> None:
        """Initialize the rule.

        Args:
            config: config for the monthly energy report rule

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if config is not valid
        """
        self._config = config
        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.energy_sum.name)
        self._mail = multi_notifier.connectors.connector_mail.Mail(config.parameter.config_mail)

        self._mail = multi_notifier.connectors.connector_mail.Mail(config.parameter.config_mail)

        if config.parameter.persistence_group_name is not None:
            # check if all energy items are in the given persistence group
            items_to_check = [config.items.energy_sum] + [item for share in config.parameter.known_energy_shares for item in share.get_items_as_list]
            not_in_persistence_group = [item.name for item in items_to_check if config.parameter.persistence_group_name not in item.groups]
            if not_in_persistence_group:
                msg = f"The following OpenHAB items are not in the persistence group '{config.parameter.persistence_group_name}': {not_in_persistence_group}"
                raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

        self.run.at(self.run.trigger.time("00:00:00").only_on(self.run.filter.days(1)), self._cb_send_energy)

        if config.parameter.debug:
            self._instance_logger.warning("Debug mode is active!")
            self.run.soon(self._cb_send_energy)
        self._instance_logger.info(f"Successfully initiated monthly consumption rule for {config.items.energy_sum.name}.")

    def _create_html(self, energy_sum_month: float) -> str:
        """Create html which will be sent by the mail.

        The template was created by https://app.bootstrapemail.com/editor/documents with the following input:

        <html>
          <head>
            <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
            <style>
            </style>
          </head>
          <body class="bg-light">
            <div class="container">
              <div class="card my-10">
                <div class="card-body">
                  <h1 class="h3 mb-2">Strom Verbrauch</h1>
                  <h5 class="text-teal-700">von Februar</h5>
                  <hr>
                  <div class="space-y-3">
                    <p class="text-gray-700">Aktueller Zählerstand: <b>7000 kWh</b>.</p>
                    <p class="text-gray-700">Hier die Details:</p>
                    <p><img src="https://www.datylon.com/hubfs/Datylon%20Website2020/Datylon%20Chart%20library/Chart%20pages/Pie%20Chart/datylon-chart-library-pie-chart-intro-example.svg" alt="Italian Trulli" align="left">
                    </p>
                  </div>
                  <hr>
                   <p style="font-size: 0.6em">Generated with habapp_rules version = 20.0.3</p>
                </div>
              </div>
            </div>
          </body>
        </html>

        Args:
            energy_sum_month: sum value for the current month

        Returns:
            html with replaced values
        """
        with (pathlib.Path(__file__).parent / "monthly_report_template.html").open(encoding="utf-8") as template_file:
            html_template = template_file.read()

        return jinja2.Template(html_template).render(
            month=_get_previous_month_name(),
            energy_now=f"{self._config.items.energy_sum.value:.1f}",
            energy_last_month=f"{energy_sum_month:.1f}",
            habapp_version=habapp_rules.__version__,
            chart="{{ chart }}",  # this is needed to not replace the chart from the mail-template
        )

    def _cb_send_energy(self) -> None:
        """Send the mail with the energy consumption of the last month."""
        self._instance_logger.debug("Send energy consumption was triggered.")
        # get values
        now = datetime.datetime.now()
        last_month = now - dateutil.relativedelta.relativedelta(months=1)

        energy_sum_month = self._config.items.energy_sum.value - habapp_rules.energy.helper.get_historic_value(self._config.items.energy_sum, last_month)
        for share in self._config.parameter.known_energy_shares:
            share.monthly_power = share.get_energy_since(last_month)

        energy_unknown = energy_sum_month - sum(share.monthly_power for share in self._config.parameter.known_energy_shares)

        with tempfile.TemporaryDirectory() as temp_dir_name:
            # create plot
            labels = [share.chart_name for share in self._config.parameter.known_energy_shares] + ["Rest"]
            values = [share.monthly_power for share in self._config.parameter.known_energy_shares] + [energy_unknown]
            chart_path = pathlib.Path(temp_dir_name) / "chart.png"
            habapp_rules.energy.donut_chart.create_chart(labels, values, chart_path)

            # get html
            html = self._create_html(energy_sum_month)

            # send mail
            self._mail.send_message(self._config.parameter.recipients, html, f"Stromverbrauch {_get_previous_month_name()}", images={"chart": str(chart_path)})

        self._instance_logger.info(f"Successfully sent energy consumption mail to {self._config.parameter.recipients}.")



"""Test for hysteresis switch."""

import collections
import unittest

import habapp_rules.common.hysteresis


class TestHysteresis(unittest.TestCase):
    """Tests for HysteresisSwitch."""

    def setUp(self) -> None:
        """Setup for all test cases."""
        self.hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(42, 2)
        self.hysteresis_switch_on_off = habapp_rules.common.hysteresis.HysteresisSwitch(42, 2, False)

    def test_get_output(self) -> None:
        """Test get_output."""
        TestCase = collections.namedtuple("TestCase", "threshold, hysteresis, state, value, expected_result")

        test_cases = [
            # hysteresis = 1, current_state = False
            TestCase(10, 1, False, 9, False),
            TestCase(10, 1, False, 9.4, False),
            TestCase(10, 1, False, 9.5, False),
            TestCase(10, 1, False, 9.6, False),
            TestCase(10, 1, False, 10.4, False),
            TestCase(10, 1, False, 10.5, True),
            TestCase(10, 1, False, 10.5, True),
            # hysteresis = 1, current_state = True
            TestCase(10, 1, True, 9, False),
            TestCase(10, 1, True, 9.4, False),
            TestCase(10, 1, True, 9.5, True),
            TestCase(10, 1, True, 9.6, True),
            TestCase(10, 1, True, 10.4, True),
            TestCase(10, 1, True, 10.5, True),
            TestCase(10, 1, True, 10.5, True),
            # hysteresis = 4, current_state = False
            TestCase(42, 4, False, 39.9, False),
            TestCase(42, 4, False, 40, False),
            TestCase(42, 4, False, 40.1, False),
            TestCase(42, 4, False, 42, False),
            TestCase(42, 4, False, 43.9, False),
            TestCase(42, 4, False, 44, True),
            TestCase(42, 4, False, 44.1, True),
            # hysteresis = 4, current_state = True
            TestCase(42, 4, True, 39.9, False),
            TestCase(42, 4, True, 40, True),
            TestCase(42, 4, True, 40.1, True),
            TestCase(42, 4, True, 42, True),
            TestCase(42, 4, True, 43.9, True),
            TestCase(42, 4, True, 44, True),
            TestCase(42, 4, True, 44.1, True),
            # threshold not set -> always False
            TestCase(None, 4, True, 39.9, False),
            TestCase(None, 4, True, 40, False),
            TestCase(None, 4, True, 40.1, False),
            TestCase(None, 4, True, 42, False),
            TestCase(None, 4, True, 43.9, False),
            TestCase(None, 4, True, 44, False),
            TestCase(None, 4, True, 44.1, False),
        ]

        for test_case in test_cases:
            self.hysteresis_switch._threshold = test_case.threshold
            self.hysteresis_switch._hysteresis = test_case.hysteresis
            self.hysteresis_switch._on_off_state = test_case.state
            self.hysteresis_switch_on_off._threshold = test_case.threshold
            self.hysteresis_switch_on_off._hysteresis = test_case.hysteresis
            self.hysteresis_switch_on_off._on_off_state = test_case.state

            self.assertEqual(test_case.expected_result, self.hysteresis_switch.get_output(test_case.value))
            self.assertEqual("ON" if test_case.expected_result else "OFF", self.hysteresis_switch_on_off.get_output(test_case.value))

            self.assertEqual(test_case.expected_result, self.hysteresis_switch._on_off_state)
            self.assertEqual(test_case.expected_result, self.hysteresis_switch_on_off._on_off_state)

    def test_get_output_without_argument(self) -> None:
        """Test get_output without value argument."""
        self.hysteresis_switch._value_last = 10
        self.assertFalse(self.hysteresis_switch.get_output())

        self.hysteresis_switch._value_last = 100
        self.assertTrue(self.hysteresis_switch.get_output())

        self.hysteresis_switch.get_output(50)
        self.assertEqual(50, self.hysteresis_switch._value_last)

    def test_set_threshold(self) -> None:
        """Test set_threshold."""
        self.assertEqual(42, self.hysteresis_switch._threshold)
        self.assertEqual(42, self.hysteresis_switch_on_off._threshold)

        self.hysteresis_switch.set_threshold_on(83)
        self.hysteresis_switch_on_off.set_threshold_on(83)

        self.assertEqual(83, self.hysteresis_switch._threshold)
        self.assertEqual(83, self.hysteresis_switch_on_off._threshold)



"""Tests for donut chart."""

import unittest
import unittest.mock

import habapp_rules.energy.donut_chart


class TestDonutFunctions(unittest.TestCase):
    """Test all donut plot functions."""

    def test_auto_percent_format(self) -> None:
        """Test _auto_percent_format."""
        values = [100, 20, 80]
        percent_values = [val / sum(values) * 100 for val in values]

        label_function = habapp_rules.energy.donut_chart._auto_percent_format(values)

        for idx, percent_value in enumerate(percent_values):
            self.assertEqual(f"{values[idx]:.1f} kWh", label_function(percent_value))

    def test_create_chart(self) -> None:
        """Test create_chart."""
        labels = ["one", "two", "three"]
        values = [1, 2, 3.0]
        path = unittest.mock.MagicMock()

        with unittest.mock.patch("habapp_rules.energy.donut_chart.plt") as pyplot_mock:
            ax_mock = unittest.mock.MagicMock()
            pyplot_mock.subplots.return_value = None, ax_mock
            text_mock_1 = unittest.mock.MagicMock()
            text_mock_2 = unittest.mock.MagicMock()
            ax_mock.pie.return_value = None, [text_mock_1, text_mock_2], None

            habapp_rules.energy.donut_chart.create_chart(labels, values, path)

        pyplot_mock.subplots.assert_called_once()
        ax_mock.pie.assert_called_once_with(values, labels=labels, autopct=unittest.mock.ANY, pctdistance=0.7, textprops={"fontsize": 10})

        text_mock_1.set_backgroundcolor.assert_called_once_with("white")
        text_mock_2.set_backgroundcolor.assert_called_once_with("white")

        pyplot_mock.savefig.assert_called_once_with(str(path), bbox_inches="tight", transparent=True)



import datetime

import HABApp

import habapp_rules.system.config.task
from habapp_rules.core.helper import send_if_different


class RecurringTask(HABApp.Rule):
    """Rule to check and set recurring tasks.

    # Items:
    Switch    Task        "Task"
    DateTime  Task_last   "Task last done".

    # Config:
    config = habapp_rules.system.config.task.RecurringTaskConfig(
        items=habapp_rules.system.config.task.RecurringTaskItems(
            task_active="I99_99_ToDo_1",
            last_done="I99_99_ToDo_1_last"
        ),
        parameter=habapp_rules.system.config.task.RecurringTaskParameter(
            recurrence_time=datetime.timedelta(hours=12))
    ))

    # Rule init:
    habapp_rules.system.task.RecurringTask(config)
    """

    def __init__(self, config: habapp_rules.system.config.task.RecurringTaskConfig) -> None:
        """Init rule.

        Args:
            config: config for this rule
        """
        HABApp.Rule.__init__(self)
        self._config = config

        if self._config.parameter.fixed_check_time is not None:
            self.run.at(self.run.trigger.time(self._config.parameter.fixed_check_time), self._check_and_set_task_undone)
        else:
            self.run.at(self.run.trigger.interval(1, self._get_check_cycle()), self._check_and_set_task_undone)

        self._config.items.task_active.listen_event(self._cb_task_active, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _get_check_cycle(self) -> datetime.timedelta:
        """Get cycle time to check if task is done.

        Returns:
            cycle time
        """
        return self._config.parameter.recurrence_time / 20

    def _cb_task_active(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if the "task_active" item was changed.

        Args:
            event: event, which triggered this callback
        """
        if event.value == "OFF":
            self._config.items.last_done.oh_send_command(datetime.datetime.now())

    def _check_and_set_task_undone(self) -> None:
        """Check if task should be set to True."""
        last_done_time = self._config.items.last_done.value if self._config.items.last_done.value is not None else datetime.datetime.min.replace()

        if last_done_time + self._config.parameter.recurrence_time < datetime.datetime.now():
            send_if_different(self._config.items.task_active, "ON")
